function refractiveindex_air(T=Float64; printBool = true)
	printBool && println("The refractive index of air is assumed to be 1 and wavelength independent.")
	f(λ) = one(Complex{T})
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_fusedsilica(T=Float64; printBool = true)
	#References:
	# I. H. Malitson. Interspecimen comparison of the refractive index of fused silica, J. Opt. Soc. Am. 55, 1205-1208 (1965)
	# C. Z. Tan. Determination of refractive index of silica glass for infrared wavelengths by IR spectroscopy, J. Non-Cryst. Solids 223, 158-163 (1998)

	printBool && println("The formula implemented for fused silica refractive index is only valid for wavelength between 0.21μm and 6.7μm")
	f(λ) = √(complex(0.691663 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.0684043^2) + 0.4079426 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.1162414^2) + 0.8974794 * (λ*1E6)^2 / ((λ*1E6)^2 - 9.896161^2) + 1))
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_zirconia(T=Float64; printBool = true)
	# Reference:
	# D. L Wood and K. Nassau. Refractive index of cubic zirconia stabilized with yttria, Appl. Opt. 21, 2978-2981 (1982)
	#
	printBool && println("The formula implemented for zirconia refractive index is only valid for wavelength between 0.361μm and 5.135μm.");
	T = Float64
	f(λ) = √(complex(1.347091 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.062543^2) + 2.117788 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.166739^2) + 9.452943 * (λ*1E6)^2 / ((λ*1E6)^2 - 24.320570^2) + 1))
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_cryolite(T = Float64; printBool = true)
	# No dispersive data found. Value from wikipedia

	printBool && println("The value for cryolite refractive index is non dispersive.");
	T = Float64
	f(λ) = one(Complex{T}) * T(1.330)
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_zincsulfide(T = Float64; printBool = true)
	# References:
	# M. Debenham. Refractive indices of zinc sulfide in the 0.405-13-µm wavelength range, Appl. Opt., 23, 2238-2239 (1984)
	# C. A. Klein. Room-temperature dispersion equations for cubic zinc sulfide, Appl. Opt. 25, 1873-1875 (1986)

	printBool && println("The formula implemented for zinc sulfide refractive index is only valid for wavelength between 0.405m and 13μm.");
	T = Float64
	f(λ) = √(complex(8.393 + .14383 / ((λ*1E6)^2 - 0.2421^2) + 4430.99 / ((λ*1E6)^2 - 36.71^2)));
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_bk7(T = Float64; printBool = true)
	# References:
	# SCHOTT Zemax catalog 2017-01-20b (obtained from http://www.schott.com)

	printBool && println("The formula implemented for BK7 glass refractive index is only valid for wavelength between 0.3μm and 2.5μm.")
	T = Float64
	f(λ) = √(complex(1 + 1.03961212 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.00600069867) + 0.231792344 * (λ*1E6)^2 / ((λ*1E6)^2 - 0.0200179144) + 1.01046945 * (λ*1E6)^2 / ((λ*1E6)^2 - 103.560653)))
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_quartz(T = Float64; printBool = true)
	# Reference:
	# G. Ghosh. Dispersion-equation coefficients for the refractive index and birefringence of calcite and quartz crystals, Opt. Commun. 163, 95-102 (1999)

	printBool && println("The formula implemented for quartz refractive index is only valid for wavelength between 0.198μm and 2.0531μm.");
	T = Float64
	f(λ) = √(complex(1 + 0.28604141 + 1.07044083 * (λ*1E6)^2 / ((λ*1E6)^2 - .0100585997) + 1.10202242 * (λ*1E6)^2 / ((λ*1E6)^2 - 100)))
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_pmma(T = Float64; printBool = true)
	# Reference:
	# Real part
	# G. Beadie, M. Brindza, R. A. Flynn, A. Rosenberg, and J. S. Shirk. Refractive index measurements of poly(methyl methacrylate) (PMMA) from 0.4-1.6 μm, Appl. Opt. 54, F139-F143 (2015)
	# Imaginary part
	# Measure experimentally on a block of PMMA 1 cm thick by Dylan Marques on Perkin Elmer Spectrophotometer.

	printBool && println("The formula implemented for real PMMA refractive index is only valid for wavelength between 0.42 μm and 1.62 μm while the data from the imaginary part is only valid for wavelength between 1 μm and 2 μm.")
	T = Float64
	n(λ) = √(complex(2.1778 + 6.1209E-3 * (λ*1E6)^2 - 1.5004E-3 * (λ*1E6)^4 + 2.3678E-2 * (λ*1E6)^(-2) - 4.2137E-3 * (λ*1E6)^(-4) + 7.3417E-4 * (λ*1E6)^(-6) - 4.5042E-5 * (λ*1E6)^(-8)));

	λ = range(1400E-9, step = 5E-9, stop = 1700E-9)
	k = 1E-4 .* [0.0939; 0.0884; 0.0877; 0.0903; 0.0939; 0.0945; 0.0905; 0.0843; 0.0808; 0.0725; 0.0572; 0.0463; 0.0424; 0.0401; 0.0376; 0.0351; 0.0332; 0.0303; 0.0271; 0.0251; 0.0241; 0.0226; 0.0207; 0.0195; 0.0190; 0.0186; 0.0184; 0.0186; 0.0191; 0.0200; 0.0213; 0.0226; 0.0245; 0.0270; 0.0289; 0.0308; 0.0330; 0.0355; 0.0386; 0.0422; 0.0464; 0.0518; 0.0591; 0.0696; 0.0829; 0.0881; 0.0941; 0.1039; 0.1151; 0.1408; 0.1972; 0.3062; 0.4575; 0.5812; 0.6690; 0.8270; 0.8990; 0.7600; 0.6439; 0.6517; 0.6916];

	kitp = interpolate(k, BSpline(Linear()))
	kitp = scale(kitp, λ)
	kitp = extrapolate(kitp, Flat());

	f(λ) = n(λ) + kitp(λ) * im
	return JolabFunction1D{T,Complex{T}}(f)
end

function refractiveindex_parylen(T = Float64; printBool = true)
	# Reference:
	# To be completed
	#
	printBool && println("The data for parylen refractive index is only valid for wavelength between 0.4 μm and 2.0 μm.");
	T = Float64

	λrefn = [4.000000e-07; 4.010000e-07; 4.020000e-07; 4.030000e-07; 4.040000e-07; 4.050000e-07; 4.060000e-07; 4.070000e-07; 4.080000e-07; 4.090000e-07; 4.100000e-07; 4.110000e-07; 4.120000e-07; 4.130000e-07; 4.140000e-07; 4.150000e-07; 4.160000e-07; 4.170000e-07; 4.180000e-07; 4.190000e-07; 4.200000e-07; 4.210000e-07; 4.220000e-07; 4.230000e-07; 4.240000e-07; 4.250000e-07; 4.260000e-07; 4.270000e-07; 4.280000e-07; 4.290000e-07; 4.300000e-07; 4.310000e-07; 4.320000e-07; 4.330000e-07; 4.340000e-07; 4.350000e-07; 4.360000e-07; 4.370000e-07; 4.380000e-07; 4.390000e-07; 4.400000e-07; 4.410000e-07; 4.420000e-07; 4.430000e-07; 4.440000e-07; 4.450000e-07; 4.460000e-07; 4.470000e-07; 4.480000e-07; 4.490000e-07; 4.500000e-07; 4.510000e-07; 4.520000e-07; 4.530000e-07; 4.540000e-07; 4.550000e-07; 4.560000e-07; 4.570000e-07; 4.580000e-07; 4.590000e-07; 4.600000e-07; 4.610000e-07; 4.620000e-07; 4.630000e-07; 4.640000e-07; 4.650000e-07; 4.660000e-07; 4.670000e-07; 4.680000e-07; 4.690000e-07; 4.700000e-07; 4.710000e-07; 4.720000e-07; 4.730000e-07; 4.740000e-07; 4.750000e-07; 4.760000e-07; 4.770000e-07; 4.780000e-07; 4.790000e-07; 4.800000e-07; 4.810000e-07; 4.820000e-07; 4.830000e-07; 4.840000e-07; 4.850000e-07; 4.860000e-07; 4.870000e-07; 4.880000e-07; 4.890000e-07; 4.900000e-07; 4.910000e-07; 4.920000e-07; 4.930000e-07; 4.940000e-07; 4.950000e-07; 4.960000e-07; 4.970000e-07; 4.980000e-07; 4.990000e-07; 5.000000e-07; 5.010000e-07; 5.020000e-07; 5.030000e-07; 5.040000e-07; 5.050000e-07; 5.060000e-07; 5.070000e-07; 5.080000e-07; 5.090000e-07; 5.100000e-07; 5.110000e-07; 5.120000e-07; 5.130000e-07; 5.140000e-07; 5.150000e-07; 5.160000e-07; 5.170000e-07; 5.180000e-07; 5.190000e-07; 5.200000e-07; 5.210000e-07; 5.220000e-07; 5.230000e-07; 5.240000e-07; 5.250000e-07; 5.260000e-07; 5.270000e-07; 5.280000e-07; 5.290000e-07; 5.300000e-07; 5.310000e-07; 5.320000e-07; 5.330000e-07; 5.340000e-07; 5.350000e-07; 5.360000e-07; 5.370000e-07; 5.380000e-07; 5.390000e-07; 5.400000e-07; 5.410000e-07; 5.420000e-07; 5.430000e-07; 5.440000e-07; 5.450000e-07; 5.460000e-07; 5.470000e-07; 5.480000e-07; 5.490000e-07; 5.500000e-07; 5.510000e-07; 5.520000e-07; 5.530000e-07; 5.540000e-07; 5.550000e-07; 5.560000e-07; 5.570000e-07; 5.580000e-07; 5.590000e-07; 5.600000e-07; 5.610000e-07; 5.620000e-07; 5.630000e-07; 5.640000e-07; 5.650000e-07; 5.660000e-07; 5.670000e-07; 5.680000e-07; 5.690000e-07; 5.700000e-07; 5.710000e-07; 5.720000e-07; 5.730000e-07; 5.740000e-07; 5.750000e-07; 5.760000e-07; 5.770000e-07; 5.780000e-07; 5.790000e-07; 5.800000e-07; 5.810000e-07; 5.820000e-07; 5.830000e-07; 5.840000e-07; 5.850000e-07; 5.860000e-07; 5.870000e-07; 5.880000e-07; 5.890000e-07; 5.900000e-07; 5.910000e-07; 5.920000e-07; 5.930000e-07; 5.940000e-07; 5.950000e-07; 5.960000e-07; 5.970000e-07; 5.980000e-07; 5.990000e-07; 6.000000e-07; 6.010000e-07; 6.020000e-07; 6.030000e-07; 6.040000e-07; 6.050000e-07; 6.060000e-07; 6.070000e-07; 6.080000e-07; 6.090000e-07; 6.100000e-07; 6.110000e-07; 6.120000e-07; 6.130000e-07; 6.140000e-07; 6.150000e-07; 6.160000e-07; 6.170000e-07; 6.180000e-07; 6.190000e-07; 6.200000e-07; 6.210000e-07; 6.220000e-07; 6.230000e-07; 6.240000e-07; 6.250000e-07; 6.260000e-07; 6.270000e-07; 6.280000e-07; 6.290000e-07; 6.300000e-07; 6.310000e-07; 6.320000e-07; 6.330000e-07; 6.340000e-07; 6.350000e-07; 6.360000e-07; 6.370000e-07; 6.380000e-07; 6.390000e-07; 6.400000e-07; 6.410000e-07; 6.420000e-07; 6.430000e-07; 6.440000e-07; 6.450000e-07; 6.460000e-07; 6.470000e-07; 6.480000e-07; 6.490000e-07; 6.500000e-07; 6.510000e-07; 6.520000e-07; 6.530000e-07; 6.540000e-07; 6.550000e-07; 6.560000e-07; 6.570000e-07; 6.580000e-07; 6.590000e-07; 6.600000e-07; 6.610000e-07; 6.620000e-07; 6.630000e-07; 6.640000e-07; 6.650000e-07; 6.660000e-07; 6.670000e-07; 6.680000e-07; 6.690000e-07; 6.700000e-07; 6.710000e-07; 6.720000e-07; 6.730000e-07; 6.740000e-07; 6.750000e-07; 6.760000e-07; 6.770000e-07; 6.780000e-07; 6.790000e-07; 6.800000e-07; 6.810000e-07; 6.820000e-07; 6.830000e-07; 6.840000e-07; 6.850000e-07; 6.860000e-07; 6.870000e-07; 6.880000e-07; 6.890000e-07; 6.900000e-07; 6.910000e-07; 6.920000e-07; 6.930000e-07; 6.940000e-07; 6.950000e-07; 6.960000e-07; 6.970000e-07; 6.980000e-07; 6.990000e-07; 7.000000e-07; 7.010000e-07; 7.020000e-07; 7.030000e-07; 7.040000e-07; 7.050000e-07; 7.060000e-07; 7.070000e-07; 7.080000e-07; 7.090000e-07; 7.100000e-07; 7.110000e-07; 7.120000e-07; 7.130000e-07; 7.140000e-07; 7.150000e-07; 7.160000e-07; 7.170000e-07; 7.180000e-07; 7.190000e-07; 7.200000e-07; 7.210000e-07; 7.220000e-07; 7.230000e-07; 7.240000e-07; 7.250000e-07; 7.260000e-07; 7.270000e-07; 7.280000e-07; 7.290000e-07; 7.300000e-07; 7.310000e-07; 7.320000e-07; 7.330000e-07; 7.340000e-07; 7.350000e-07; 7.360000e-07; 7.370000e-07; 7.380000e-07; 7.390000e-07; 7.400000e-07; 7.410000e-07; 7.420000e-07; 7.430000e-07; 7.440000e-07; 7.450000e-07; 7.460000e-07; 7.470000e-07; 7.480000e-07; 7.490000e-07; 7.500000e-07; 7.510000e-07; 7.520000e-07; 7.530000e-07; 7.540000e-07; 7.550000e-07; 7.560000e-07; 7.570000e-07; 7.580000e-07; 7.590000e-07; 7.600000e-07; 7.610000e-07; 7.620000e-07; 7.630000e-07; 7.640000e-07; 7.650000e-07; 7.660000e-07; 7.670000e-07; 7.680000e-07; 7.690000e-07; 7.700000e-07; 7.710000e-07; 7.720000e-07; 7.730000e-07; 7.740000e-07; 7.750000e-07; 7.760000e-07; 7.770000e-07; 7.780000e-07; 7.790000e-07; 7.800000e-07; 7.810000e-07; 7.820000e-07; 7.830000e-07; 7.840000e-07; 7.850000e-07; 7.860000e-07; 7.870000e-07; 7.880000e-07; 7.890000e-07; 7.900000e-07; 7.910000e-07; 7.920000e-07; 7.930000e-07; 7.940000e-07; 7.950000e-07; 7.960000e-07; 7.970000e-07; 7.980000e-07; 7.990000e-07; 8.000000e-07; 8.010000e-07; 8.020000e-07; 8.030000e-07; 8.040000e-07; 8.050000e-07; 8.060000e-07; 8.070000e-07; 8.080000e-07; 8.090000e-07; 8.100000e-07; 8.110000e-07; 8.120000e-07; 8.130000e-07; 8.140000e-07; 8.150000e-07; 8.160000e-07; 8.170000e-07; 8.180000e-07; 8.190000e-07; 8.200000e-07; 8.210000e-07; 8.220000e-07; 8.230000e-07; 8.240000e-07; 8.250000e-07; 8.260000e-07; 8.270000e-07; 8.280000e-07; 8.290000e-07; 8.300000e-07; 8.310000e-07; 8.320000e-07; 8.330000e-07; 8.340000e-07; 8.350000e-07; 8.360000e-07; 8.370000e-07; 8.380000e-07; 8.390000e-07; 8.400000e-07; 8.410000e-07; 8.420000e-07; 8.430000e-07; 8.440000e-07; 8.450000e-07; 8.460000e-07; 8.470000e-07; 8.480000e-07; 8.490000e-07; 8.500000e-07; 8.510000e-07; 8.520000e-07; 8.530000e-07; 8.540000e-07; 8.550000e-07; 8.560000e-07; 8.570000e-07; 8.580000e-07; 8.590000e-07; 8.600000e-07; 8.610000e-07; 8.620000e-07; 8.630000e-07; 8.640000e-07; 8.650000e-07; 8.660000e-07; 8.670000e-07; 8.680000e-07; 8.690000e-07; 8.700000e-07; 8.710000e-07; 8.720000e-07; 8.730000e-07; 8.740000e-07; 8.750000e-07; 8.760000e-07; 8.770000e-07; 8.780000e-07; 8.790000e-07; 8.800000e-07; 8.810000e-07; 8.820000e-07; 8.830000e-07; 8.840000e-07; 8.850000e-07; 8.860000e-07; 8.870000e-07; 8.880000e-07; 8.890000e-07; 8.900000e-07; 8.910000e-07; 8.920000e-07; 8.930000e-07; 8.940000e-07; 8.950000e-07; 8.960000e-07; 8.970000e-07; 8.980000e-07; 8.990000e-07; 9.000000e-07; 9.010000e-07; 9.020000e-07; 9.030000e-07; 9.040000e-07; 9.050000e-07; 9.060000e-07; 9.070000e-07; 9.080000e-07; 9.090000e-07; 9.100000e-07; 9.110000e-07; 9.120000e-07; 9.130000e-07; 9.140000e-07; 9.150000e-07; 9.160000e-07; 9.170000e-07; 9.180000e-07; 9.190000e-07; 9.200000e-07; 9.210000e-07; 9.220000e-07; 9.230000e-07; 9.240000e-07; 9.250000e-07; 9.260000e-07; 9.270000e-07; 9.280000e-07; 9.290000e-07; 9.300000e-07; 9.310000e-07; 9.320000e-07; 9.330000e-07; 9.340000e-07; 9.350000e-07; 9.360000e-07; 9.370000e-07; 9.380000e-07; 9.390000e-07; 9.400000e-07; 9.410000e-07; 9.420000e-07; 9.430000e-07; 9.440000e-07; 9.450000e-07; 9.460000e-07; 9.470000e-07; 9.480000e-07; 9.490000e-07; 9.500000e-07; 9.510000e-07; 9.520000e-07; 9.530000e-07; 9.540000e-07; 9.550000e-07; 9.560000e-07; 9.570000e-07; 9.580000e-07; 9.590000e-07; 9.600000e-07; 9.610000e-07; 9.620000e-07; 9.630000e-07; 9.640000e-07; 9.650000e-07; 9.660000e-07; 9.670000e-07; 9.680000e-07; 9.690000e-07; 9.700000e-07; 9.710000e-07; 9.720000e-07; 9.730000e-07; 9.740000e-07; 9.750000e-07; 9.760000e-07; 9.770000e-07; 9.780000e-07; 9.790000e-07; 9.800000e-07; 9.810000e-07; 9.820000e-07; 9.830000e-07; 9.840000e-07; 9.850000e-07; 9.860000e-07; 9.870000e-07; 9.880000e-07; 9.890000e-07; 9.900000e-07; 9.910000e-07; 9.920000e-07; 9.930000e-07; 9.940000e-07; 9.950000e-07; 9.960000e-07; 9.970000e-07; 9.980000e-07; 9.990000e-07; 1.000000e-06; 1.001000e-06; 1.002000e-06; 1.003000e-06; 1.004000e-06; 1.005000e-06; 1.006000e-06; 1.007000e-06; 1.008000e-06; 1.009000e-06; 1.010000e-06; 1.011000e-06; 1.012000e-06; 1.013000e-06; 1.014000e-06; 1.015000e-06; 1.016000e-06; 1.017000e-06; 1.018000e-06; 1.019000e-06; 1.020000e-06; 1.021000e-06; 1.022000e-06; 1.023000e-06; 1.024000e-06; 1.025000e-06; 1.026000e-06; 1.027000e-06; 1.028000e-06; 1.029000e-06; 1.030000e-06; 1.031000e-06; 1.032000e-06; 1.033000e-06; 1.034000e-06; 1.035000e-06; 1.036000e-06; 1.037000e-06; 1.038000e-06; 1.039000e-06; 1.040000e-06; 1.041000e-06; 1.042000e-06; 1.043000e-06; 1.044000e-06; 1.045000e-06; 1.046000e-06; 1.047000e-06; 1.048000e-06; 1.049000e-06; 1.050000e-06; 1.051000e-06; 1.052000e-06; 1.053000e-06; 1.054000e-06; 1.055000e-06; 1.056000e-06; 1.057000e-06; 1.058000e-06; 1.059000e-06; 1.060000e-06; 1.061000e-06; 1.062000e-06; 1.063000e-06; 1.064000e-06; 1.065000e-06; 1.066000e-06; 1.067000e-06; 1.068000e-06; 1.069000e-06; 1.070000e-06; 1.071000e-06; 1.072000e-06; 1.073000e-06; 1.074000e-06; 1.075000e-06; 1.076000e-06; 1.077000e-06; 1.078000e-06; 1.079000e-06; 1.080000e-06; 1.081000e-06; 1.082000e-06; 1.083000e-06; 1.084000e-06; 1.085000e-06; 1.086000e-06; 1.087000e-06; 1.088000e-06; 1.089000e-06; 1.090000e-06; 1.091000e-06; 1.092000e-06; 1.093000e-06; 1.094000e-06; 1.095000e-06; 1.096000e-06; 1.097000e-06; 1.098000e-06; 1.099000e-06; 1.100000e-06; 1.101000e-06; 1.102000e-06; 1.103000e-06; 1.104000e-06; 1.105000e-06; 1.106000e-06; 1.107000e-06; 1.108000e-06; 1.109000e-06; 1.110000e-06; 1.111000e-06; 1.112000e-06; 1.113000e-06; 1.114000e-06; 1.115000e-06; 1.116000e-06; 1.117000e-06; 1.118000e-06; 1.119000e-06; 1.120000e-06; 1.121000e-06; 1.122000e-06; 1.123000e-06; 1.124000e-06; 1.125000e-06; 1.126000e-06; 1.127000e-06; 1.128000e-06; 1.129000e-06; 1.130000e-06; 1.131000e-06; 1.132000e-06; 1.133000e-06; 1.134000e-06; 1.135000e-06; 1.136000e-06; 1.137000e-06; 1.138000e-06; 1.139000e-06; 1.140000e-06; 1.141000e-06; 1.142000e-06; 1.143000e-06; 1.144000e-06; 1.145000e-06; 1.146000e-06; 1.147000e-06; 1.148000e-06; 1.149000e-06; 1.150000e-06; 1.151000e-06; 1.152000e-06; 1.153000e-06; 1.154000e-06; 1.155000e-06; 1.156000e-06; 1.157000e-06; 1.158000e-06; 1.159000e-06; 1.160000e-06; 1.161000e-06; 1.162000e-06; 1.163000e-06; 1.164000e-06; 1.165000e-06; 1.166000e-06; 1.167000e-06; 1.168000e-06; 1.169000e-06; 1.170000e-06; 1.171000e-06; 1.172000e-06; 1.173000e-06; 1.174000e-06; 1.175000e-06; 1.176000e-06; 1.177000e-06; 1.178000e-06; 1.179000e-06; 1.180000e-06; 1.181000e-06; 1.182000e-06; 1.183000e-06; 1.184000e-06; 1.185000e-06; 1.186000e-06; 1.187000e-06; 1.188000e-06; 1.189000e-06; 1.190000e-06; 1.191000e-06; 1.192000e-06; 1.193000e-06; 1.194000e-06; 1.195000e-06; 1.196000e-06; 1.197000e-06; 1.198000e-06; 1.199000e-06; 1.200000e-06; 1.201000e-06; 1.202000e-06; 1.203000e-06; 1.204000e-06; 1.205000e-06; 1.206000e-06; 1.207000e-06; 1.208000e-06; 1.209000e-06; 1.210000e-06; 1.211000e-06; 1.212000e-06; 1.213000e-06; 1.214000e-06; 1.215000e-06; 1.216000e-06; 1.217000e-06; 1.218000e-06; 1.219000e-06; 1.220000e-06; 1.221000e-06; 1.222000e-06; 1.223000e-06; 1.224000e-06; 1.225000e-06; 1.226000e-06; 1.227000e-06; 1.228000e-06; 1.229000e-06; 1.230000e-06; 1.231000e-06; 1.232000e-06; 1.233000e-06; 1.234000e-06; 1.235000e-06; 1.236000e-06; 1.237000e-06; 1.238000e-06; 1.239000e-06; 1.240000e-06; 1.241000e-06; 1.242000e-06; 1.243000e-06; 1.244000e-06; 1.245000e-06; 1.246000e-06; 1.247000e-06; 1.248000e-06; 1.249000e-06; 1.250000e-06; 1.251000e-06; 1.252000e-06; 1.253000e-06; 1.254000e-06; 1.255000e-06; 1.256000e-06; 1.257000e-06; 1.258000e-06; 1.259000e-06; 1.260000e-06; 1.261000e-06; 1.262000e-06; 1.263000e-06; 1.264000e-06; 1.265000e-06; 1.266000e-06; 1.267000e-06; 1.268000e-06; 1.269000e-06; 1.270000e-06; 1.271000e-06; 1.272000e-06; 1.273000e-06; 1.274000e-06; 1.275000e-06; 1.276000e-06; 1.277000e-06; 1.278000e-06; 1.279000e-06; 1.280000e-06; 1.281000e-06; 1.282000e-06; 1.283000e-06; 1.284000e-06; 1.285000e-06; 1.286000e-06; 1.287000e-06; 1.288000e-06; 1.289000e-06; 1.290000e-06; 1.291000e-06; 1.292000e-06; 1.293000e-06; 1.294000e-06; 1.295000e-06; 1.296000e-06; 1.297000e-06; 1.298000e-06; 1.299000e-06; 1.300000e-06; 1.301000e-06; 1.302000e-06; 1.303000e-06; 1.304000e-06; 1.305000e-06; 1.306000e-06; 1.307000e-06; 1.308000e-06; 1.309000e-06; 1.310000e-06; 1.311000e-06; 1.312000e-06; 1.313000e-06; 1.314000e-06; 1.315000e-06; 1.316000e-06; 1.317000e-06; 1.318000e-06; 1.319000e-06; 1.320000e-06; 1.321000e-06; 1.322000e-06; 1.323000e-06; 1.324000e-06; 1.325000e-06; 1.326000e-06; 1.327000e-06; 1.328000e-06; 1.329000e-06; 1.330000e-06; 1.331000e-06; 1.332000e-06; 1.333000e-06; 1.334000e-06; 1.335000e-06; 1.336000e-06; 1.337000e-06; 1.338000e-06; 1.339000e-06; 1.340000e-06; 1.341000e-06; 1.342000e-06; 1.343000e-06; 1.344000e-06; 1.345000e-06; 1.346000e-06; 1.347000e-06; 1.348000e-06; 1.349000e-06; 1.350000e-06; 1.351000e-06; 1.352000e-06; 1.353000e-06; 1.354000e-06; 1.355000e-06; 1.356000e-06; 1.357000e-06; 1.358000e-06; 1.359000e-06; 1.360000e-06; 1.361000e-06; 1.362000e-06; 1.363000e-06; 1.364000e-06; 1.365000e-06; 1.366000e-06; 1.367000e-06; 1.368000e-06; 1.369000e-06; 1.370000e-06; 1.371000e-06; 1.372000e-06; 1.373000e-06; 1.374000e-06; 1.375000e-06; 1.376000e-06; 1.377000e-06; 1.378000e-06; 1.379000e-06; 1.380000e-06; 1.381000e-06; 1.382000e-06; 1.383000e-06; 1.384000e-06; 1.385000e-06; 1.386000e-06; 1.387000e-06; 1.388000e-06; 1.389000e-06; 1.390000e-06; 1.391000e-06; 1.392000e-06; 1.393000e-06; 1.394000e-06; 1.395000e-06; 1.396000e-06; 1.397000e-06; 1.398000e-06; 1.399000e-06; 1.400000e-06; 1.401000e-06; 1.402000e-06; 1.403000e-06; 1.404000e-06; 1.405000e-06; 1.406000e-06; 1.407000e-06; 1.408000e-06; 1.409000e-06; 1.410000e-06; 1.411000e-06; 1.412000e-06; 1.413000e-06; 1.414000e-06; 1.415000e-06; 1.416000e-06; 1.417000e-06; 1.418000e-06; 1.419000e-06; 1.420000e-06; 1.421000e-06; 1.422000e-06; 1.423000e-06; 1.424000e-06; 1.425000e-06; 1.426000e-06; 1.427000e-06; 1.428000e-06; 1.429000e-06; 1.430000e-06; 1.431000e-06; 1.432000e-06; 1.433000e-06; 1.434000e-06; 1.435000e-06; 1.436000e-06; 1.437000e-06; 1.438000e-06; 1.439000e-06; 1.440000e-06; 1.441000e-06; 1.442000e-06; 1.443000e-06; 1.444000e-06; 1.445000e-06; 1.446000e-06; 1.447000e-06; 1.448000e-06; 1.449000e-06; 1.450000e-06; 1.451000e-06; 1.452000e-06; 1.453000e-06; 1.454000e-06; 1.455000e-06; 1.456000e-06; 1.457000e-06; 1.458000e-06; 1.459000e-06; 1.460000e-06; 1.461000e-06; 1.462000e-06; 1.463000e-06; 1.464000e-06; 1.465000e-06; 1.466000e-06; 1.467000e-06; 1.468000e-06; 1.469000e-06; 1.470000e-06; 1.471000e-06; 1.472000e-06; 1.473000e-06; 1.474000e-06; 1.475000e-06; 1.476000e-06; 1.477000e-06; 1.478000e-06; 1.479000e-06; 1.480000e-06; 1.481000e-06; 1.482000e-06; 1.483000e-06; 1.484000e-06; 1.485000e-06; 1.486000e-06; 1.487000e-06; 1.488000e-06; 1.489000e-06; 1.490000e-06; 1.491000e-06; 1.492000e-06; 1.493000e-06; 1.494000e-06; 1.495000e-06; 1.496000e-06; 1.497000e-06; 1.498000e-06; 1.499000e-06; 1.500000e-06; 1.501000e-06; 1.502000e-06; 1.503000e-06; 1.504000e-06; 1.505000e-06; 1.506000e-06; 1.507000e-06; 1.508000e-06; 1.509000e-06; 1.510000e-06; 1.511000e-06; 1.512000e-06; 1.513000e-06; 1.514000e-06; 1.515000e-06; 1.516000e-06; 1.517000e-06; 1.518000e-06; 1.519000e-06; 1.520000e-06; 1.521000e-06; 1.522000e-06; 1.523000e-06; 1.524000e-06; 1.525000e-06; 1.526000e-06; 1.527000e-06; 1.528000e-06; 1.529000e-06; 1.530000e-06; 1.531000e-06; 1.532000e-06; 1.533000e-06; 1.534000e-06; 1.535000e-06; 1.536000e-06; 1.537000e-06; 1.538000e-06; 1.539000e-06; 1.540000e-06; 1.541000e-06; 1.542000e-06; 1.543000e-06; 1.544000e-06; 1.545000e-06; 1.546000e-06; 1.547000e-06; 1.548000e-06; 1.549000e-06; 1.550000e-06; 1.551000e-06; 1.552000e-06; 1.553000e-06; 1.554000e-06; 1.555000e-06; 1.556000e-06; 1.557000e-06; 1.558000e-06; 1.559000e-06; 1.560000e-06; 1.561000e-06; 1.562000e-06; 1.563000e-06; 1.564000e-06; 1.565000e-06; 1.566000e-06; 1.567000e-06; 1.568000e-06; 1.569000e-06; 1.570000e-06; 1.571000e-06; 1.572000e-06; 1.573000e-06; 1.574000e-06; 1.575000e-06; 1.576000e-06; 1.577000e-06; 1.578000e-06; 1.579000e-06; 1.580000e-06; 1.581000e-06; 1.582000e-06; 1.583000e-06; 1.584000e-06; 1.585000e-06; 1.586000e-06; 1.587000e-06; 1.588000e-06; 1.589000e-06; 1.590000e-06; 1.591000e-06; 1.592000e-06; 1.593000e-06; 1.594000e-06; 1.595000e-06; 1.596000e-06; 1.597000e-06; 1.598000e-06; 1.599000e-06; 1.600000e-06; 1.601000e-06; 1.602000e-06; 1.603000e-06; 1.604000e-06; 1.605000e-06; 1.606000e-06; 1.607000e-06; 1.608000e-06; 1.609000e-06; 1.610000e-06; 1.611000e-06; 1.612000e-06; 1.613000e-06; 1.614000e-06; 1.615000e-06; 1.616000e-06; 1.617000e-06; 1.618000e-06; 1.619000e-06; 1.620000e-06; 1.621000e-06; 1.622000e-06; 1.623000e-06; 1.624000e-06; 1.625000e-06; 1.626000e-06; 1.627000e-06; 1.628000e-06; 1.629000e-06; 1.630000e-06; 1.631000e-06; 1.632000e-06; 1.633000e-06; 1.634000e-06; 1.635000e-06; 1.636000e-06; 1.637000e-06; 1.638000e-06; 1.639000e-06; 1.640000e-06; 1.641000e-06; 1.642000e-06; 1.643000e-06; 1.644000e-06; 1.645000e-06; 1.646000e-06; 1.647000e-06; 1.648000e-06; 1.649000e-06; 1.650000e-06; 1.651000e-06; 1.652000e-06; 1.653000e-06; 1.654000e-06; 1.655000e-06; 1.656000e-06; 1.657000e-06; 1.658000e-06; 1.659000e-06; 1.660000e-06; 1.661000e-06; 1.662000e-06; 1.663000e-06; 1.664000e-06; 1.665000e-06; 1.666000e-06; 1.667000e-06; 1.668000e-06; 1.669000e-06; 1.670000e-06; 1.671000e-06; 1.672000e-06; 1.673000e-06; 1.674000e-06; 1.675000e-06; 1.676000e-06; 1.677000e-06; 1.678000e-06; 1.679000e-06; 1.680000e-06; 1.681000e-06; 1.682000e-06; 1.683000e-06; 1.684000e-06; 1.685000e-06; 1.686000e-06; 1.687000e-06; 1.688000e-06; 1.689000e-06; 1.690000e-06; 1.691000e-06; 1.692000e-06; 1.693000e-06; 1.694000e-06; 1.695000e-06; 1.696000e-06; 1.697000e-06; 1.698000e-06; 1.699000e-06; 1.700000e-06; 1.701000e-06; 1.702000e-06; 1.703000e-06; 1.704000e-06; 1.705000e-06; 1.706000e-06; 1.707000e-06; 1.708000e-06; 1.709000e-06; 1.710000e-06; 1.711000e-06; 1.712000e-06; 1.713000e-06; 1.714000e-06; 1.715000e-06; 1.716000e-06; 1.717000e-06; 1.718000e-06; 1.719000e-06; 1.720000e-06; 1.721000e-06; 1.722000e-06; 1.723000e-06; 1.724000e-06; 1.725000e-06; 1.726000e-06; 1.727000e-06; 1.728000e-06; 1.729000e-06; 1.730000e-06; 1.731000e-06; 1.732000e-06; 1.733000e-06; 1.734000e-06; 1.735000e-06; 1.736000e-06; 1.737000e-06; 1.738000e-06; 1.739000e-06; 1.740000e-06; 1.741000e-06; 1.742000e-06; 1.743000e-06; 1.744000e-06; 1.745000e-06; 1.746000e-06; 1.747000e-06; 1.748000e-06; 1.749000e-06; 1.750000e-06; 1.751000e-06; 1.752000e-06; 1.753000e-06; 1.754000e-06; 1.755000e-06; 1.756000e-06; 1.757000e-06; 1.758000e-06; 1.759000e-06; 1.760000e-06; 1.761000e-06; 1.762000e-06; 1.763000e-06; 1.764000e-06; 1.765000e-06; 1.766000e-06; 1.767000e-06; 1.768000e-06; 1.769000e-06; 1.770000e-06; 1.771000e-06; 1.772000e-06; 1.773000e-06; 1.774000e-06; 1.775000e-06; 1.776000e-06; 1.777000e-06; 1.778000e-06; 1.779000e-06; 1.780000e-06; 1.781000e-06; 1.782000e-06; 1.783000e-06; 1.784000e-06; 1.785000e-06; 1.786000e-06; 1.787000e-06; 1.788000e-06; 1.789000e-06; 1.790000e-06; 1.791000e-06; 1.792000e-06; 1.793000e-06; 1.794000e-06; 1.795000e-06; 1.796000e-06; 1.797000e-06; 1.798000e-06; 1.799000e-06; 1.800000e-06; 1.801000e-06; 1.802000e-06; 1.803000e-06; 1.804000e-06; 1.805000e-06; 1.806000e-06; 1.807000e-06; 1.808000e-06; 1.809000e-06; 1.810000e-06; 1.811000e-06; 1.812000e-06; 1.813000e-06; 1.814000e-06; 1.815000e-06; 1.816000e-06; 1.817000e-06; 1.818000e-06; 1.819000e-06; 1.820000e-06; 1.821000e-06; 1.822000e-06; 1.823000e-06; 1.824000e-06; 1.825000e-06; 1.826000e-06; 1.827000e-06; 1.828000e-06; 1.829000e-06; 1.830000e-06; 1.831000e-06; 1.832000e-06; 1.833000e-06; 1.834000e-06; 1.835000e-06; 1.836000e-06; 1.837000e-06; 1.838000e-06; 1.839000e-06; 1.840000e-06; 1.841000e-06; 1.842000e-06; 1.843000e-06; 1.844000e-06; 1.845000e-06; 1.846000e-06; 1.847000e-06; 1.848000e-06; 1.849000e-06; 1.850000e-06; 1.851000e-06; 1.852000e-06; 1.853000e-06; 1.854000e-06; 1.855000e-06; 1.856000e-06; 1.857000e-06; 1.858000e-06; 1.859000e-06; 1.860000e-06; 1.861000e-06; 1.862000e-06; 1.863000e-06; 1.864000e-06; 1.865000e-06; 1.866000e-06; 1.867000e-06; 1.868000e-06; 1.869000e-06; 1.870000e-06; 1.871000e-06; 1.872000e-06; 1.873000e-06; 1.874000e-06; 1.875000e-06; 1.876000e-06; 1.877000e-06; 1.878000e-06; 1.879000e-06; 1.880000e-06; 1.881000e-06; 1.882000e-06; 1.883000e-06; 1.884000e-06; 1.885000e-06; 1.886000e-06; 1.887000e-06; 1.888000e-06; 1.889000e-06; 1.890000e-06; 1.891000e-06; 1.892000e-06; 1.893000e-06; 1.894000e-06; 1.895000e-06; 1.896000e-06; 1.897000e-06; 1.898000e-06; 1.899000e-06; 1.900000e-06; 1.901000e-06; 1.902000e-06; 1.903000e-06; 1.904000e-06; 1.905000e-06; 1.906000e-06; 1.907000e-06; 1.908000e-06; 1.909000e-06; 1.910000e-06; 1.911000e-06; 1.912000e-06; 1.913000e-06; 1.914000e-06; 1.915000e-06; 1.916000e-06; 1.917000e-06; 1.918000e-06; 1.919000e-06; 1.920000e-06; 1.921000e-06; 1.922000e-06; 1.923000e-06; 1.924000e-06; 1.925000e-06; 1.926000e-06; 1.927000e-06; 1.928000e-06; 1.929000e-06; 1.930000e-06; 1.931000e-06; 1.932000e-06; 1.933000e-06; 1.934000e-06; 1.935000e-06; 1.936000e-06; 1.937000e-06; 1.938000e-06; 1.939000e-06; 1.940000e-06; 1.941000e-06; 1.942000e-06; 1.943000e-06; 1.944000e-06; 1.945000e-06; 1.946000e-06; 1.947000e-06; 1.948000e-06; 1.949000e-06; 1.950000e-06; 1.951000e-06; 1.952000e-06; 1.953000e-06; 1.954000e-06; 1.955000e-06; 1.956000e-06; 1.957000e-06; 1.958000e-06; 1.959000e-06; 1.960000e-06; 1.961000e-06; 1.962000e-06; 1.963000e-06; 1.964000e-06; 1.965000e-06; 1.966000e-06; 1.967000e-06; 1.968000e-06; 1.969000e-06; 1.970000e-06; 1.971000e-06; 1.972000e-06; 1.973000e-06; 1.974000e-06; 1.975000e-06; 1.976000e-06; 1.977000e-06; 1.978000e-06; 1.979000e-06; 1.980000e-06; 1.981000e-06; 1.982000e-06; 1.983000e-06; 1.984000e-06; 1.985000e-06; 1.986000e-06; 1.987000e-06; 1.988000e-06; 1.989000e-06; 1.990000e-06; 1.991000e-06; 1.992000e-06; 1.993000e-06; 1.994000e-06; 1.995000e-06; 1.996000e-06; 1.997000e-06; 1.998000e-06; 1.999000e-06; 2.000000e-06];
	nref = [1.665625; 1.665323; 1.665023; 1.664726; 1.664430; 1.664137; 1.663846; 1.663558; 1.663271; 1.662986; 1.662704; 1.662423; 1.662145; 1.661868; 1.661594; 1.661322; 1.661051; 1.660783; 1.660516; 1.660251; 1.659989; 1.659728; 1.659469; 1.659211; 1.658956; 1.658702; 1.658451; 1.658201; 1.657952; 1.657706; 1.657461; 1.657218; 1.656976; 1.656736; 1.656498; 1.656262; 1.656027; 1.655794; 1.655562; 1.655332; 1.655103; 1.654876; 1.654651; 1.654427; 1.654205; 1.653984; 1.653764; 1.653546; 1.653330; 1.653115; 1.652901; 1.652689; 1.652478; 1.652269; 1.652061; 1.651854; 1.651649; 1.651445; 1.651242; 1.651041; 1.650841; 1.650643; 1.650445; 1.650249; 1.650054; 1.649861; 1.649668; 1.649477; 1.649287; 1.649099; 1.648911; 1.648725; 1.648540; 1.648356; 1.648173; 1.647992; 1.647811; 1.647632; 1.647454; 1.647277; 1.647101; 1.646926; 1.646752; 1.646579; 1.646408; 1.646237; 1.646068; 1.645899; 1.645732; 1.645565; 1.645400; 1.645235; 1.645072; 1.644910; 1.644748; 1.644588; 1.644428; 1.644270; 1.644112; 1.643956; 1.643800; 1.643645; 1.643491; 1.643339; 1.643187; 1.643035; 1.642885; 1.642736; 1.642588; 1.642440; 1.642293; 1.642148; 1.642003; 1.641858; 1.641715; 1.641573; 1.641431; 1.641290; 1.641150; 1.641011; 1.640873; 1.640735; 1.640598; 1.640462; 1.640327; 1.640193; 1.640059; 1.639926; 1.639794; 1.639663; 1.639532; 1.639402; 1.639273; 1.639144; 1.639016; 1.638889; 1.638763; 1.638637; 1.638513; 1.638388; 1.638265; 1.638142; 1.638020; 1.637898; 1.637777; 1.637657; 1.637538; 1.637419; 1.637301; 1.637183; 1.637066; 1.636950; 1.636834; 1.636719; 1.636605; 1.636491; 1.636378; 1.636265; 1.636153; 1.636042; 1.635931; 1.635821; 1.635711; 1.635602; 1.635494; 1.635386; 1.635279; 1.635172; 1.635066; 1.634960; 1.634855; 1.634751; 1.634647; 1.634544; 1.634441; 1.634338; 1.634237; 1.634135; 1.634035; 1.633934; 1.633835; 1.633736; 1.633637; 1.633539; 1.633441; 1.633344; 1.633247; 1.633151; 1.633055; 1.632960; 1.632866; 1.632771; 1.632678; 1.632584; 1.632492; 1.632399; 1.632307; 1.632216; 1.632125; 1.632034; 1.631944; 1.631855; 1.631766; 1.631677; 1.631589; 1.631501; 1.631414; 1.631327; 1.631240; 1.631154; 1.631068; 1.630983; 1.630898; 1.630814; 1.630730; 1.630646; 1.630563; 1.630480; 1.630398; 1.630316; 1.630234; 1.630153; 1.630072; 1.629992; 1.629912; 1.629832; 1.629753; 1.629674; 1.629595; 1.629517; 1.629439; 1.629362; 1.629285; 1.629208; 1.629132; 1.629056; 1.628980; 1.628905; 1.628830; 1.628756; 1.628682; 1.628608; 1.628534; 1.628461; 1.628388; 1.628316; 1.628244; 1.628172; 1.628101; 1.628029; 1.627959; 1.627888; 1.627818; 1.627748; 1.627679; 1.627609; 1.627541; 1.627472; 1.627404; 1.627336; 1.627268; 1.627201; 1.627134; 1.627067; 1.627001; 1.626935; 1.626869; 1.626803; 1.626738; 1.626673; 1.626608; 1.626544; 1.626480; 1.626416; 1.626353; 1.626289; 1.626226; 1.626164; 1.626101; 1.626039; 1.625978; 1.625916; 1.625855; 1.625794; 1.625733; 1.625672; 1.625612; 1.625552; 1.625492; 1.625433; 1.625374; 1.625315; 1.625256; 1.625198; 1.625140; 1.625082; 1.625024; 1.624967; 1.624910; 1.624853; 1.624796; 1.624739; 1.624683; 1.624627; 1.624572; 1.624516; 1.624461; 1.624406; 1.624351; 1.624297; 1.624242; 1.624188; 1.624134; 1.624081; 1.624027; 1.623974; 1.623921; 1.623868; 1.623816; 1.623764; 1.623711; 1.623660; 1.623608; 1.623556; 1.623505; 1.623454; 1.623403; 1.623353; 1.623302; 1.623252; 1.623202; 1.623153; 1.623103; 1.623054; 1.623004; 1.622955; 1.622907; 1.622858; 1.622810; 1.622762; 1.622714; 1.622666; 1.622618; 1.622571; 1.622524; 1.622477; 1.622430; 1.622383; 1.622337; 1.622291; 1.622244; 1.622199; 1.622153; 1.622107; 1.622062; 1.622017; 1.621972; 1.621927; 1.621882; 1.621838; 1.621794; 1.621750; 1.621706; 1.621662; 1.621618; 1.621575; 1.621532; 1.621488; 1.621446; 1.621403; 1.621360; 1.621318; 1.621276; 1.621234; 1.621192; 1.621150; 1.621108; 1.621067; 1.621026; 1.620984; 1.620943; 1.620903; 1.620862; 1.620822; 1.620781; 1.620741; 1.620701; 1.620661; 1.620621; 1.620582; 1.620542; 1.620503; 1.620464; 1.620425; 1.620386; 1.620347; 1.620309; 1.620271; 1.620232; 1.620194; 1.620156; 1.620118; 1.620081; 1.620043; 1.620006; 1.619969; 1.619931; 1.619894; 1.619858; 1.619821; 1.619784; 1.619748; 1.619712; 1.619675; 1.619639; 1.619603; 1.619568; 1.619532; 1.619497; 1.619461; 1.619426; 1.619391; 1.619356; 1.619321; 1.619286; 1.619252; 1.619217; 1.619183; 1.619149; 1.619114; 1.619080; 1.619047; 1.619013; 1.618979; 1.618946; 1.618912; 1.618879; 1.618846; 1.618813; 1.618780; 1.618747; 1.618714; 1.618682; 1.618649; 1.618617; 1.618585; 1.618553; 1.618521; 1.618489; 1.618457; 1.618426; 1.618394; 1.618363; 1.618331; 1.618300; 1.618269; 1.618238; 1.618207; 1.618176; 1.618146; 1.618115; 1.618085; 1.618054; 1.618024; 1.617994; 1.617964; 1.617934; 1.617904; 1.617875; 1.617845; 1.617815; 1.617786; 1.617757; 1.617728; 1.617698; 1.617669; 1.617640; 1.617612; 1.617583; 1.617554; 1.617526; 1.617497; 1.617469; 1.617441; 1.617413; 1.617385; 1.617357; 1.617329; 1.617301; 1.617273; 1.617246; 1.617218; 1.617191; 1.617164; 1.617137; 1.617109; 1.617082; 1.617056; 1.617029; 1.617002; 1.616975; 1.616949; 1.616922; 1.616896; 1.616870; 1.616843; 1.616817; 1.616791; 1.616765; 1.616739; 1.616714; 1.616688; 1.616662; 1.616637; 1.616611; 1.616586; 1.616561; 1.616535; 1.616510; 1.616485; 1.616460; 1.616435; 1.616411; 1.616386; 1.616361; 1.616337; 1.616312; 1.616288; 1.616264; 1.616239; 1.616215; 1.616191; 1.616167; 1.616143; 1.616119; 1.616096; 1.616072; 1.616048; 1.616025; 1.616001; 1.615978; 1.615954; 1.615931; 1.615908; 1.615885; 1.615862; 1.615839; 1.615816; 1.615793; 1.615771; 1.615748; 1.615725; 1.615703; 1.615680; 1.615658; 1.615636; 1.615613; 1.615591; 1.615569; 1.615547; 1.615525; 1.615503; 1.615481; 1.615460; 1.615438; 1.615416; 1.615395; 1.615373; 1.615352; 1.615331; 1.615309; 1.615288; 1.615267; 1.615246; 1.615225; 1.615204; 1.615183; 1.615162; 1.615141; 1.615121; 1.615100; 1.615079; 1.615059; 1.615038; 1.615018; 1.614998; 1.614977; 1.614957; 1.614937; 1.614917; 1.614897; 1.614877; 1.614857; 1.614837; 1.614817; 1.614798; 1.614778; 1.614758; 1.614739; 1.614719; 1.614700; 1.614681; 1.614661; 1.614642; 1.614623; 1.614604; 1.614585; 1.614566; 1.614547; 1.614528; 1.614509; 1.614490; 1.614471; 1.614453; 1.614434; 1.614415; 1.614397; 1.614378; 1.614360; 1.614342; 1.614323; 1.614305; 1.614287; 1.614269; 1.614251; 1.614233; 1.614215; 1.614197; 1.614179; 1.614161; 1.614143; 1.614125; 1.614108; 1.614090; 1.614073; 1.614055; 1.614038; 1.614020; 1.614003; 1.613985; 1.613968; 1.613951; 1.613934; 1.613917; 1.613900; 1.613883; 1.613866; 1.613849; 1.613832; 1.613815; 1.613798; 1.613781; 1.613765; 1.613748; 1.613732; 1.613715; 1.613698; 1.613682; 1.613666; 1.613649; 1.613633; 1.613617; 1.613600; 1.613584; 1.613568; 1.613552; 1.613536; 1.613520; 1.613504; 1.613488; 1.613472; 1.613457; 1.613441; 1.613425; 1.613409; 1.613394; 1.613378; 1.613363; 1.613347; 1.613332; 1.613316; 1.613301; 1.613285; 1.613270; 1.613255; 1.613240; 1.613225; 1.613209; 1.613194; 1.613179; 1.613164; 1.613149; 1.613134; 1.613120; 1.613105; 1.613090; 1.613075; 1.613060; 1.613046; 1.613031; 1.613017; 1.613002; 1.612987; 1.612973; 1.612959; 1.612944; 1.612930; 1.612915; 1.612901; 1.612887; 1.612873; 1.612859; 1.612844; 1.612830; 1.612816; 1.612802; 1.612788; 1.612774; 1.612760; 1.612747; 1.612733; 1.612719; 1.612705; 1.612692; 1.612678; 1.612664; 1.612651; 1.612637; 1.612623; 1.612610; 1.612597; 1.612583; 1.612570; 1.612556; 1.612543; 1.612530; 1.612516; 1.612503; 1.612490; 1.612477; 1.612464; 1.612451; 1.612438; 1.612425; 1.612412; 1.612399; 1.612386; 1.612373; 1.612360; 1.612347; 1.612335; 1.612322; 1.612309; 1.612296; 1.612284; 1.612271; 1.612259; 1.612246; 1.612234; 1.612221; 1.612209; 1.612196; 1.612184; 1.612172; 1.612159; 1.612147; 1.612135; 1.612122; 1.612110; 1.612098; 1.612086; 1.612074; 1.612062; 1.612050; 1.612038; 1.612026; 1.612014; 1.612002; 1.611990; 1.611978; 1.611966; 1.611955; 1.611943; 1.611931; 1.611919; 1.611908; 1.611896; 1.611884; 1.611873; 1.611861; 1.611850; 1.611838; 1.611827; 1.611815; 1.611804; 1.611793; 1.611781; 1.611770; 1.611759; 1.611747; 1.611736; 1.611725; 1.611714; 1.611703; 1.611691; 1.611680; 1.611669; 1.611658; 1.611647; 1.611636; 1.611625; 1.611614; 1.611603; 1.611592; 1.611582; 1.611571; 1.611560; 1.611549; 1.611538; 1.611528; 1.611517; 1.611506; 1.611496; 1.611485; 1.611475; 1.611464; 1.611453; 1.611443; 1.611432; 1.611422; 1.611412; 1.611401; 1.611391; 1.611380; 1.611370; 1.611360; 1.611349; 1.611339; 1.611329; 1.611319; 1.611309; 1.611298; 1.611288; 1.611278; 1.611268; 1.611258; 1.611248; 1.611238; 1.611228; 1.611218; 1.611208; 1.611198; 1.611188; 1.611178; 1.611168; 1.611159; 1.611149; 1.611139; 1.611129; 1.611120; 1.611110; 1.611100; 1.611091; 1.611081; 1.611071; 1.611062; 1.611052; 1.611043; 1.611033; 1.611023; 1.611014; 1.611005; 1.610995; 1.610986; 1.610976; 1.610967; 1.610958; 1.610948; 1.610939; 1.610930; 1.610920; 1.610911; 1.610902; 1.610893; 1.610884; 1.610874; 1.610865; 1.610856; 1.610847; 1.610838; 1.610829; 1.610820; 1.610811; 1.610802; 1.610793; 1.610784; 1.610775; 1.610766; 1.610757; 1.610748; 1.610740; 1.610731; 1.610722; 1.610713; 1.610704; 1.610696; 1.610687; 1.610678; 1.610670; 1.610661; 1.610652; 1.610644; 1.610635; 1.610627; 1.610618; 1.610609; 1.610601; 1.610592; 1.610584; 1.610575; 1.610567; 1.610559; 1.610550; 1.610542; 1.610533; 1.610525; 1.610517; 1.610508; 1.610500; 1.610492; 1.610484; 1.610475; 1.610467; 1.610459; 1.610451; 1.610443; 1.610434; 1.610426; 1.610418; 1.610410; 1.610402; 1.610394; 1.610386; 1.610378; 1.610370; 1.610362; 1.610354; 1.610346; 1.610338; 1.610330; 1.610322; 1.610314; 1.610307; 1.610299; 1.610291; 1.610283; 1.610275; 1.610268; 1.610260; 1.610252; 1.610244; 1.610237; 1.610229; 1.610221; 1.610214; 1.610206; 1.610198; 1.610191; 1.610183; 1.610176; 1.610168; 1.610161; 1.610153; 1.610146; 1.610138; 1.610131; 1.610123; 1.610116; 1.610108; 1.610101; 1.610093; 1.610086; 1.610079; 1.610071; 1.610064; 1.610057; 1.610049; 1.610042; 1.610035; 1.610028; 1.610020; 1.610013; 1.610006; 1.609999; 1.609992; 1.609985; 1.609977; 1.609970; 1.609963; 1.609956; 1.609949; 1.609942; 1.609935; 1.609928; 1.609921; 1.609914; 1.609907; 1.609900; 1.609893; 1.609886; 1.609879; 1.609872; 1.609865; 1.609858; 1.609851; 1.609845; 1.609838; 1.609831; 1.609824; 1.609817; 1.609811; 1.609804; 1.609797; 1.609790; 1.609784; 1.609777; 1.609770; 1.609763; 1.609757; 1.609750; 1.609744; 1.609737; 1.609730; 1.609724; 1.609717; 1.609711; 1.609704; 1.609697; 1.609691; 1.609684; 1.609678; 1.609671; 1.609665; 1.609658; 1.609652; 1.609646; 1.609639; 1.609633; 1.609626; 1.609620; 1.609614; 1.609607; 1.609601; 1.609595; 1.609588; 1.609582; 1.609576; 1.609569; 1.609563; 1.609557; 1.609551; 1.609544; 1.609538; 1.609532; 1.609526; 1.609520; 1.609513; 1.609507; 1.609501; 1.609495; 1.609489; 1.609483; 1.609477; 1.609471; 1.609465; 1.609458; 1.609452; 1.609446; 1.609440; 1.609434; 1.609428; 1.609422; 1.609416; 1.609411; 1.609405; 1.609399; 1.609393; 1.609387; 1.609381; 1.609375; 1.609369; 1.609363; 1.609357; 1.609352; 1.609346; 1.609340; 1.609334; 1.609328; 1.609323; 1.609317; 1.609311; 1.609305; 1.609300; 1.609294; 1.609288; 1.609283; 1.609277; 1.609271; 1.609265; 1.609260; 1.609254; 1.609249; 1.609243; 1.609237; 1.609232; 1.609226; 1.609221; 1.609215; 1.609209; 1.609204; 1.609198; 1.609193; 1.609187; 1.609182; 1.609176; 1.609171; 1.609165; 1.609160; 1.609155; 1.609149; 1.609144; 1.609138; 1.609133; 1.609128; 1.609122; 1.609117; 1.609111; 1.609106; 1.609101; 1.609095; 1.609090; 1.609085; 1.609079; 1.609074; 1.609069; 1.609064; 1.609058; 1.609053; 1.609048; 1.609043; 1.609037; 1.609032; 1.609027; 1.609022; 1.609017; 1.609012; 1.609006; 1.609001; 1.608996; 1.608991; 1.608986; 1.608981; 1.608976; 1.608971; 1.608966; 1.608960; 1.608955; 1.608950; 1.608945; 1.608940; 1.608935; 1.608930; 1.608925; 1.608920; 1.608915; 1.608910; 1.608905; 1.608900; 1.608895; 1.608891; 1.608886; 1.608881; 1.608876; 1.608871; 1.608866; 1.608861; 1.608856; 1.608851; 1.608847; 1.608842; 1.608837; 1.608832; 1.608827; 1.608822; 1.608818; 1.608813; 1.608808; 1.608803; 1.608799; 1.608794; 1.608789; 1.608784; 1.608780; 1.608775; 1.608770; 1.608765; 1.608761; 1.608756; 1.608751; 1.608747; 1.608742; 1.608737; 1.608733; 1.608728; 1.608724; 1.608719; 1.608714; 1.608710; 1.608705; 1.608701; 1.608696; 1.608692; 1.608687; 1.608682; 1.608678; 1.608673; 1.608669; 1.608664; 1.608660; 1.608655; 1.608651; 1.608646; 1.608642; 1.608637; 1.608633; 1.608629; 1.608624; 1.608620; 1.608615; 1.608611; 1.608606; 1.608602; 1.608598; 1.608593; 1.608589; 1.608585; 1.608580; 1.608576; 1.608572; 1.608567; 1.608563; 1.608559; 1.608554; 1.608550; 1.608546; 1.608541; 1.608537; 1.608533; 1.608529; 1.608524; 1.608520; 1.608516; 1.608512; 1.608507; 1.608503; 1.608499; 1.608495; 1.608491; 1.608486; 1.608482; 1.608478; 1.608474; 1.608470; 1.608466; 1.608461; 1.608457; 1.608453; 1.608449; 1.608445; 1.608441; 1.608437; 1.608433; 1.608429; 1.608425; 1.608420; 1.608416; 1.608412; 1.608408; 1.608404; 1.608400; 1.608396; 1.608392; 1.608388; 1.608384; 1.608380; 1.608376; 1.608372; 1.608368; 1.608364; 1.608360; 1.608356; 1.608352; 1.608349; 1.608345; 1.608341; 1.608337; 1.608333; 1.608329; 1.608325; 1.608321; 1.608317; 1.608313; 1.608310; 1.608306; 1.608302; 1.608298; 1.608294; 1.608290; 1.608286; 1.608283; 1.608279; 1.608275; 1.608271; 1.608267; 1.608264; 1.608260; 1.608256; 1.608252; 1.608249; 1.608245; 1.608241; 1.608237; 1.608234; 1.608230; 1.608226; 1.608222; 1.608219; 1.608215; 1.608211; 1.608208; 1.608204; 1.608200; 1.608197; 1.608193; 1.608189; 1.608186; 1.608182; 1.608178; 1.608175; 1.608171; 1.608167; 1.608164; 1.608160; 1.608157; 1.608153; 1.608149; 1.608146; 1.608142; 1.608139; 1.608135; 1.608131; 1.608128; 1.608124; 1.608121; 1.608117; 1.608114; 1.608110; 1.608107; 1.608103; 1.608100; 1.608096; 1.608093; 1.608089; 1.608086; 1.608082; 1.608079; 1.608075; 1.608072; 1.608068; 1.608065; 1.608061; 1.608058; 1.608055; 1.608051; 1.608048; 1.608044; 1.608041; 1.608038; 1.608034; 1.608031; 1.608027; 1.608024; 1.608021; 1.608017; 1.608014; 1.608011; 1.608007; 1.608004; 1.608000; 1.607997; 1.607994; 1.607991; 1.607987; 1.607984; 1.607981; 1.607977; 1.607974; 1.607971; 1.607967; 1.607964; 1.607961; 1.607958; 1.607954; 1.607951; 1.607948; 1.607945; 1.607941; 1.607938; 1.607935; 1.607932; 1.607928; 1.607925; 1.607922; 1.607919; 1.607916; 1.607912; 1.607909; 1.607906; 1.607903; 1.607900; 1.607896; 1.607893; 1.607890; 1.607887; 1.607884; 1.607881; 1.607878; 1.607874; 1.607871; 1.607868; 1.607865; 1.607862; 1.607859; 1.607856; 1.607853; 1.607850; 1.607846; 1.607843; 1.607840; 1.607837; 1.607834; 1.607831; 1.607828; 1.607825; 1.607822; 1.607819; 1.607816; 1.607813; 1.607810; 1.607807; 1.607804; 1.607801; 1.607798; 1.607795; 1.607792; 1.607789; 1.607786; 1.607783; 1.607780; 1.607777; 1.607774; 1.607771; 1.607768; 1.607765; 1.607762; 1.607759; 1.607756; 1.607753; 1.607750; 1.607747; 1.607744; 1.607742; 1.607739; 1.607736; 1.607733; 1.607730; 1.607727; 1.607724; 1.607721; 1.607718; 1.607715; 1.607713; 1.607710; 1.607707; 1.607704; 1.607701; 1.607698; 1.607695; 1.607693; 1.607690; 1.607687; 1.607684; 1.607681; 1.607679; 1.607676; 1.607673; 1.607670; 1.607667; 1.607664; 1.607662; 1.607659; 1.607656; 1.607653; 1.607651; 1.607648; 1.607645; 1.607642; 1.607640; 1.607637; 1.607634; 1.607631; 1.607629; 1.607626; 1.607623; 1.607620; 1.607618; 1.607615; 1.607612; 1.607610; 1.607607; 1.607604; 1.607601; 1.607599; 1.607596; 1.607593; 1.607591; 1.607588; 1.607585; 1.607583; 1.607580; 1.607577; 1.607575; 1.607572; 1.607569; 1.607567; 1.607564; 1.607561; 1.607559; 1.607556; 1.607554; 1.607551; 1.607548; 1.607546; 1.607543; 1.607541; 1.607538; 1.607535; 1.607533; 1.607530; 1.607528; 1.607525; 1.607522; 1.607520; 1.607517; 1.607515; 1.607512; 1.607510; 1.607507; 1.607505; 1.607502; 1.607499; 1.607497; 1.607494; 1.607492; 1.607489; 1.607487; 1.607484; 1.607482; 1.607479; 1.607477; 1.607474; 1.607472; 1.607469; 1.607467; 1.607464; 1.607462; 1.607459; 1.607457; 1.607454; 1.607452; 1.607449; 1.607447; 1.607445; 1.607442; 1.607440; 1.607437; 1.607435; 1.607432; 1.607430; 1.607427; 1.607425];

	λrefk = [3.0000000e-07; 3.0100000e-07; 3.0200000e-07; 3.0300000e-07; 3.0400000e-07; 3.0500000e-07; 3.0600000e-07; 3.0700000e-07; 3.0800000e-07; 3.0900000e-07; 3.1000000e-07; 3.1100000e-07; 3.1200000e-07; 3.1300000e-07; 3.1400000e-07; 3.1500000e-07; 3.1600000e-07; 3.1700000e-07; 3.1800000e-07; 3.1900000e-07; 3.2000000e-07; 3.2100000e-07; 3.2200000e-07; 3.2300000e-07; 3.2400000e-07; 3.2500000e-07; 3.2600000e-07; 3.2700000e-07; 3.2800000e-07; 3.2900000e-07; 3.3000000e-07; 3.3100000e-07; 3.3200000e-07; 3.3300000e-07; 3.3400000e-07; 3.3500000e-07; 3.3600000e-07; 3.3700000e-07; 3.3800000e-07; 3.3900000e-07; 3.4000000e-07; 3.4100000e-07; 3.4200000e-07; 3.4300000e-07; 3.4400000e-07; 3.4500000e-07; 3.4600000e-07; 3.4700000e-07; 3.4800000e-07; 3.4900000e-07; 3.5000000e-07; 3.5100000e-07; 3.5200000e-07; 3.5300000e-07; 3.5400000e-07; 3.5500000e-07; 3.5600000e-07; 3.5700000e-07; 3.5800000e-07; 3.5900000e-07; 3.6000000e-07; 3.6100000e-07; 3.6200000e-07; 3.6300000e-07; 3.6400000e-07; 3.6500000e-07; 3.6600000e-07; 3.6700000e-07; 3.6800000e-07; 3.6900000e-07; 3.7000000e-07; 3.7100000e-07; 3.7200000e-07; 3.7300000e-07; 3.7400000e-07; 3.7500000e-07; 3.7600000e-07; 3.7700000e-07; 3.7800000e-07; 3.7900000e-07; 3.8000000e-07; 3.8100000e-07; 3.8200000e-07; 3.8300000e-07; 3.8400000e-07; 3.8500000e-07; 3.8600000e-07; 3.8700000e-07; 3.8800000e-07; 3.8900000e-07; 3.9000000e-07; 3.9100000e-07; 3.9200000e-07; 3.9300000e-07; 3.9400000e-07; 3.9500000e-07; 3.9600000e-07; 3.9700000e-07; 3.9800000e-07; 3.9900000e-07; 4.0000000e-07; 4.0100000e-07; 4.0200000e-07; 4.0300000e-07; 4.0400000e-07; 4.0500000e-07; 4.0600000e-07; 4.0700000e-07; 4.0800000e-07; 4.0900000e-07; 4.1000000e-07; 4.1100000e-07; 4.1200000e-07; 4.1300000e-07; 4.1400000e-07; 4.1500000e-07; 4.1600000e-07; 4.1700000e-07; 4.1800000e-07; 4.1900000e-07; 4.2000000e-07; 4.2100000e-07; 4.2200000e-07; 4.2300000e-07; 4.2400000e-07; 4.2500000e-07; 4.2600000e-07; 4.2700000e-07; 4.2800000e-07; 4.2900000e-07; 4.3000000e-07; 4.3100000e-07; 4.3200000e-07; 4.3300000e-07; 4.3400000e-07; 4.3500000e-07; 4.3600000e-07; 4.3700000e-07; 4.3800000e-07; 4.3900000e-07; 4.4000000e-07; 4.4100000e-07; 4.4200000e-07; 4.4300000e-07; 4.4400000e-07; 4.4500000e-07; 4.4600000e-07; 4.4700000e-07; 4.4800000e-07; 4.4900000e-07; 4.5000000e-07; 4.5100000e-07; 4.5200000e-07; 4.5300000e-07; 4.5400000e-07; 4.5500000e-07; 4.5600000e-07; 4.5700000e-07; 4.5800000e-07; 4.5900000e-07; 4.6000000e-07; 4.6100000e-07; 4.6200000e-07; 4.6300000e-07; 4.6400000e-07; 4.6500000e-07; 4.6600000e-07; 4.6700000e-07; 4.6800000e-07; 4.6900000e-07; 4.7000000e-07; 4.7100000e-07; 4.7200000e-07; 4.7300000e-07; 4.7400000e-07; 4.7500000e-07; 4.7600000e-07; 4.7700000e-07; 4.7800000e-07; 4.7900000e-07; 4.8000000e-07; 4.8100000e-07; 4.8200000e-07; 4.8300000e-07; 4.8400000e-07; 4.8500000e-07; 4.8600000e-07; 4.8700000e-07; 4.8800000e-07; 4.8900000e-07; 4.9000000e-07; 4.9100000e-07; 4.9200000e-07; 4.9300000e-07; 4.9400000e-07; 4.9500000e-07; 4.9600000e-07; 4.9700000e-07; 4.9800000e-07; 4.9900000e-07; 5.0000000e-07; 5.0100000e-07; 5.0200000e-07; 5.0300000e-07; 5.0400000e-07; 5.0500000e-07; 5.0600000e-07; 5.0700000e-07; 5.0800000e-07; 5.0900000e-07; 5.1000000e-07; 5.1100000e-07; 5.1200000e-07; 5.1300000e-07; 5.1400000e-07; 5.1500000e-07; 5.1600000e-07; 5.1700000e-07; 5.1800000e-07; 5.1900000e-07; 5.2000000e-07; 5.2100000e-07; 5.2200000e-07; 5.2300000e-07; 5.2400000e-07; 5.2500000e-07; 5.2600000e-07; 5.2700000e-07; 5.2800000e-07; 5.2900000e-07; 5.3000000e-07; 5.3100000e-07; 5.3200000e-07; 5.3300000e-07; 5.3400000e-07; 5.3500000e-07; 5.3600000e-07; 5.3700000e-07; 5.3800000e-07; 5.3900000e-07; 5.4000000e-07; 5.4100000e-07; 5.4200000e-07; 5.4300000e-07; 5.4400000e-07; 5.4500000e-07; 5.4600000e-07; 5.4700000e-07; 5.4800000e-07; 5.4900000e-07; 5.5000000e-07; 5.5100000e-07; 5.5200000e-07; 5.5300000e-07; 5.5400000e-07; 5.5500000e-07; 5.5600000e-07; 5.5700000e-07; 5.5800000e-07; 5.5900000e-07; 5.6000000e-07; 5.6100000e-07; 5.6200000e-07; 5.6300000e-07; 5.6400000e-07; 5.6500000e-07; 5.6600000e-07; 5.6700000e-07; 5.6800000e-07; 5.6900000e-07; 5.7000000e-07; 5.7100000e-07; 5.7200000e-07; 5.7300000e-07; 5.7400000e-07; 5.7500000e-07; 5.7600000e-07; 5.7700000e-07; 5.7800000e-07; 5.7900000e-07; 5.8000000e-07; 5.8100000e-07; 5.8200000e-07; 5.8300000e-07; 5.8400000e-07; 5.8500000e-07; 5.8600000e-07; 5.8700000e-07; 5.8800000e-07; 5.8900000e-07; 5.9000000e-07; 5.9100000e-07; 5.9200000e-07; 5.9300000e-07; 5.9400000e-07; 5.9500000e-07; 5.9600000e-07; 5.9700000e-07; 5.9800000e-07; 5.9900000e-07; 6.0000000e-07; 6.0100000e-07; 6.0200000e-07; 6.0300000e-07; 6.0400000e-07; 6.0500000e-07; 6.0600000e-07; 6.0700000e-07; 6.0800000e-07; 6.0900000e-07; 6.1000000e-07; 6.1100000e-07; 6.1200000e-07; 6.1300000e-07; 6.1400000e-07; 6.1500000e-07; 6.1600000e-07; 6.1700000e-07; 6.1800000e-07; 6.1900000e-07; 6.2000000e-07; 6.2100000e-07; 6.2200000e-07; 6.2300000e-07; 6.2400000e-07; 6.2500000e-07; 6.2600000e-07; 6.2700000e-07; 6.2800000e-07; 6.2900000e-07; 6.3000000e-07; 6.3100000e-07; 6.3200000e-07; 6.3300000e-07; 6.3400000e-07; 6.3500000e-07; 6.3600000e-07; 6.3700000e-07; 6.3800000e-07; 6.3900000e-07; 6.4000000e-07; 6.4100000e-07; 6.4200000e-07; 6.4300000e-07; 6.4400000e-07; 6.4500000e-07; 6.4600000e-07; 6.4700000e-07; 6.4800000e-07; 6.4900000e-07; 6.5000000e-07; 6.5100000e-07; 6.5200000e-07; 6.5300000e-07; 6.5400000e-07; 6.5500000e-07; 6.5600000e-07; 6.5700000e-07; 6.5800000e-07; 6.5900000e-07; 6.6000000e-07; 6.6100000e-07; 6.6200000e-07; 6.6300000e-07; 6.6400000e-07; 6.6500000e-07; 6.6600000e-07; 6.6700000e-07; 6.6800000e-07; 6.6900000e-07; 6.7000000e-07; 6.7100000e-07; 6.7200000e-07; 6.7300000e-07; 6.7400000e-07; 6.7500000e-07; 6.7600000e-07; 6.7700000e-07; 6.7800000e-07; 6.7900000e-07; 6.8000000e-07; 6.8100000e-07; 6.8200000e-07; 6.8300000e-07; 6.8400000e-07; 6.8500000e-07; 6.8600000e-07; 6.8700000e-07; 6.8800000e-07; 6.8900000e-07; 6.9000000e-07; 6.9100000e-07; 6.9200000e-07; 6.9300000e-07; 6.9400000e-07; 6.9500000e-07; 6.9600000e-07; 6.9700000e-07; 6.9800000e-07; 6.9900000e-07; 7.0000000e-07; 7.0100000e-07; 7.0200000e-07; 7.0300000e-07; 7.0400000e-07; 7.0500000e-07; 7.0600000e-07; 7.0700000e-07; 7.0800000e-07; 7.0900000e-07; 7.1000000e-07; 7.1100000e-07; 7.1200000e-07; 7.1300000e-07; 7.1400000e-07; 7.1500000e-07; 7.1600000e-07; 7.1700000e-07; 7.1800000e-07; 7.1900000e-07; 7.2000000e-07; 7.2100000e-07; 7.2200000e-07; 7.2300000e-07; 7.2400000e-07; 7.2500000e-07; 7.2600000e-07; 7.2700000e-07; 7.2800000e-07; 7.2900000e-07; 7.3000000e-07; 7.3100000e-07; 7.3200000e-07; 7.3300000e-07; 7.3400000e-07; 7.3500000e-07; 7.3600000e-07; 7.3700000e-07; 7.3800000e-07; 7.3900000e-07; 7.4000000e-07; 7.4100000e-07; 7.4200000e-07; 7.4300000e-07; 7.4400000e-07; 7.4500000e-07; 7.4600000e-07; 7.4700000e-07; 7.4800000e-07; 7.4900000e-07; 7.5000000e-07; 7.5100000e-07; 7.5200000e-07; 7.5300000e-07; 7.5400000e-07; 7.5500000e-07; 7.5600000e-07; 7.5700000e-07; 7.5800000e-07; 7.5900000e-07; 7.6000000e-07; 7.6100000e-07; 7.6200000e-07; 7.6300000e-07; 7.6400000e-07; 7.6500000e-07; 7.6600000e-07; 7.6700000e-07; 7.6800000e-07; 7.6900000e-07; 7.7000000e-07; 7.7100000e-07; 7.7200000e-07; 7.7300000e-07; 7.7400000e-07; 7.7500000e-07; 7.7600000e-07; 7.7700000e-07; 7.7800000e-07; 7.7900000e-07; 7.8000000e-07; 7.8100000e-07; 7.8200000e-07; 7.8300000e-07; 7.8400000e-07; 7.8500000e-07; 7.8600000e-07; 7.8700000e-07; 7.8800000e-07; 7.8900000e-07; 7.9000000e-07; 7.9100000e-07; 7.9200000e-07; 7.9300000e-07; 7.9400000e-07; 7.9500000e-07; 7.9600000e-07; 7.9700000e-07; 7.9800000e-07; 7.9900000e-07; 8.0000000e-07; 8.0100000e-07; 8.0200000e-07; 8.0300000e-07; 8.0400000e-07; 8.0500000e-07; 8.0600000e-07; 8.0700000e-07; 8.0800000e-07; 8.0900000e-07; 8.1000000e-07; 8.1100000e-07; 8.1200000e-07; 8.1300000e-07; 8.1400000e-07; 8.1500000e-07; 8.1600000e-07; 8.1700000e-07; 8.1800000e-07; 8.1900000e-07; 8.2000000e-07; 8.2100000e-07; 8.2200000e-07; 8.2300000e-07; 8.2400000e-07; 8.2500000e-07; 8.2600000e-07; 8.2700000e-07; 8.2800000e-07; 8.2900000e-07; 8.3000000e-07; 8.3100000e-07; 8.3200000e-07; 8.3300000e-07; 8.3400000e-07; 8.3500000e-07; 8.3600000e-07; 8.3700000e-07; 8.3800000e-07; 8.3900000e-07; 8.4000000e-07; 8.4100000e-07; 8.4200000e-07; 8.4300000e-07; 8.4400000e-07; 8.4500000e-07; 8.4600000e-07; 8.4700000e-07; 8.4800000e-07; 8.4900000e-07; 8.5000000e-07; 8.5100000e-07; 8.5200000e-07; 8.5300000e-07; 8.5400000e-07; 8.5500000e-07; 8.5600000e-07; 8.5700000e-07; 8.5800000e-07; 8.5900000e-07; 8.6000000e-07; 8.6100000e-07; 8.6200000e-07; 8.6300000e-07; 8.6400000e-07; 8.6500000e-07; 8.6600000e-07; 8.6700000e-07; 8.6800000e-07; 8.6900000e-07; 8.7000000e-07; 8.7100000e-07; 8.7200000e-07; 8.7300000e-07; 8.7400000e-07; 8.7500000e-07; 8.7600000e-07; 8.7700000e-07; 8.7800000e-07; 8.7900000e-07; 8.8000000e-07; 8.8100000e-07; 8.8200000e-07; 8.8300000e-07; 8.8400000e-07; 8.8500000e-07; 8.8600000e-07; 8.8700000e-07; 8.8800000e-07; 8.8900000e-07; 8.9000000e-07; 8.9100000e-07; 8.9200000e-07; 8.9300000e-07; 8.9400000e-07; 8.9500000e-07; 8.9600000e-07; 8.9700000e-07; 8.9800000e-07; 8.9900000e-07; 9.0000000e-07; 9.0100000e-07; 9.0200000e-07; 9.0300000e-07; 9.0400000e-07; 9.0500000e-07; 9.0600000e-07; 9.0700000e-07; 9.0800000e-07; 9.0900000e-07; 9.1000000e-07; 9.1100000e-07; 9.1200000e-07; 9.1300000e-07; 9.1400000e-07; 9.1500000e-07; 9.1600000e-07; 9.1700000e-07; 9.1800000e-07; 9.1900000e-07; 9.2000000e-07; 9.2100000e-07; 9.2200000e-07; 9.2300000e-07; 9.2400000e-07; 9.2500000e-07; 9.2600000e-07; 9.2700000e-07; 9.2800000e-07; 9.2900000e-07; 9.3000000e-07; 9.3100000e-07; 9.3200000e-07; 9.3300000e-07; 9.3400000e-07; 9.3500000e-07; 9.3600000e-07; 9.3700000e-07; 9.3800000e-07; 9.3900000e-07; 9.4000000e-07; 9.4100000e-07; 9.4200000e-07; 9.4300000e-07; 9.4400000e-07; 9.4500000e-07; 9.4600000e-07; 9.4700000e-07; 9.4800000e-07; 9.4900000e-07; 9.5000000e-07; 9.5100000e-07; 9.5200000e-07; 9.5300000e-07; 9.5400000e-07; 9.5500000e-07; 9.5600000e-07; 9.5700000e-07; 9.5800000e-07; 9.5900000e-07; 9.6000000e-07; 9.6100000e-07; 9.6200000e-07; 9.6300000e-07; 9.6400000e-07; 9.6500000e-07; 9.6600000e-07; 9.6700000e-07; 9.6800000e-07; 9.6900000e-07; 9.7000000e-07; 9.7100000e-07; 9.7200000e-07; 9.7300000e-07; 9.7400000e-07; 9.7500000e-07; 9.7600000e-07; 9.7700000e-07; 9.7800000e-07; 9.7900000e-07; 9.8000000e-07; 9.8100000e-07; 9.8200000e-07; 9.8300000e-07; 9.8400000e-07; 9.8500000e-07; 9.8600000e-07; 9.8700000e-07; 9.8800000e-07; 9.8900000e-07; 9.9000000e-07; 9.9100000e-07; 9.9200000e-07; 9.9300000e-07; 9.9400000e-07; 9.9500000e-07; 9.9600000e-07; 9.9700000e-07; 9.9800000e-07; 9.9900000e-07; 1.0000000e-06; 1.0010000e-06; 1.0020000e-06; 1.0030000e-06; 1.0040000e-06; 1.0050000e-06; 1.0060000e-06; 1.0070000e-06; 1.0080000e-06; 1.0090000e-06; 1.0100000e-06; 1.0110000e-06; 1.0120000e-06; 1.0130000e-06; 1.0140000e-06; 1.0150000e-06; 1.0160000e-06; 1.0170000e-06; 1.0180000e-06; 1.0190000e-06; 1.0200000e-06; 1.0210000e-06; 1.0220000e-06; 1.0230000e-06; 1.0240000e-06; 1.0250000e-06; 1.0260000e-06; 1.0270000e-06; 1.0280000e-06; 1.0290000e-06; 1.0300000e-06; 1.0310000e-06; 1.0320000e-06; 1.0330000e-06; 1.0340000e-06; 1.0350000e-06; 1.0360000e-06; 1.0370000e-06; 1.0380000e-06; 1.0390000e-06; 1.0400000e-06; 1.0410000e-06; 1.0420000e-06; 1.0430000e-06; 1.0440000e-06; 1.0450000e-06; 1.0460000e-06; 1.0470000e-06; 1.0480000e-06; 1.0490000e-06; 1.0500000e-06; 1.0510000e-06; 1.0520000e-06; 1.0530000e-06; 1.0540000e-06; 1.0550000e-06; 1.0560000e-06; 1.0570000e-06; 1.0580000e-06; 1.0590000e-06; 1.0600000e-06; 1.0610000e-06; 1.0620000e-06; 1.0630000e-06; 1.0640000e-06; 1.0650000e-06; 1.0660000e-06; 1.0670000e-06; 1.0680000e-06; 1.0690000e-06; 1.0700000e-06; 1.0710000e-06; 1.0720000e-06; 1.0730000e-06; 1.0740000e-06; 1.0750000e-06; 1.0760000e-06; 1.0770000e-06; 1.0780000e-06; 1.0790000e-06; 1.0800000e-06; 1.0810000e-06; 1.0820000e-06; 1.0830000e-06; 1.0840000e-06; 1.0850000e-06; 1.0860000e-06; 1.0870000e-06; 1.0880000e-06; 1.0890000e-06; 1.0900000e-06; 1.0910000e-06; 1.0920000e-06; 1.0930000e-06; 1.0940000e-06; 1.0950000e-06; 1.0960000e-06; 1.0970000e-06; 1.0980000e-06; 1.0990000e-06; 1.1000000e-06; 1.1010000e-06; 1.1020000e-06; 1.1030000e-06; 1.1040000e-06; 1.1050000e-06; 1.1060000e-06; 1.1070000e-06; 1.1080000e-06; 1.1090000e-06; 1.1100000e-06; 1.1110000e-06; 1.1120000e-06; 1.1130000e-06; 1.1140000e-06; 1.1150000e-06; 1.1160000e-06; 1.1170000e-06; 1.1180000e-06; 1.1190000e-06; 1.1200000e-06; 1.1210000e-06; 1.1220000e-06; 1.1230000e-06; 1.1240000e-06; 1.1250000e-06; 1.1260000e-06; 1.1270000e-06; 1.1280000e-06; 1.1290000e-06; 1.1300000e-06; 1.1310000e-06; 1.1320000e-06; 1.1330000e-06; 1.1340000e-06; 1.1350000e-06; 1.1360000e-06; 1.1370000e-06; 1.1380000e-06; 1.1390000e-06; 1.1400000e-06; 1.1410000e-06; 1.1420000e-06; 1.1430000e-06; 1.1440000e-06; 1.1450000e-06; 1.1460000e-06; 1.1470000e-06; 1.1480000e-06; 1.1490000e-06; 1.1500000e-06; 1.1510000e-06; 1.1520000e-06; 1.1530000e-06; 1.1540000e-06; 1.1550000e-06; 1.1560000e-06; 1.1570000e-06; 1.1580000e-06; 1.1590000e-06; 1.1600000e-06; 1.1610000e-06; 1.1620000e-06; 1.1630000e-06; 1.1640000e-06; 1.1650000e-06; 1.1660000e-06; 1.1670000e-06; 1.1680000e-06; 1.1690000e-06; 1.1700000e-06; 1.1710000e-06; 1.1720000e-06; 1.1730000e-06; 1.1740000e-06; 1.1750000e-06; 1.1760000e-06; 1.1770000e-06; 1.1780000e-06; 1.1790000e-06; 1.1800000e-06; 1.1810000e-06; 1.1820000e-06; 1.1830000e-06; 1.1840000e-06; 1.1850000e-06; 1.1860000e-06; 1.1870000e-06; 1.1880000e-06; 1.1890000e-06; 1.1900000e-06; 1.1910000e-06; 1.1920000e-06; 1.1930000e-06; 1.1940000e-06; 1.1950000e-06; 1.1960000e-06; 1.1970000e-06; 1.1980000e-06; 1.1990000e-06; 1.2000000e-06; 1.2010000e-06; 1.2020000e-06; 1.2030000e-06; 1.2040000e-06; 1.2050000e-06; 1.2060000e-06; 1.2070000e-06; 1.2080000e-06; 1.2090000e-06; 1.2100000e-06; 1.2110000e-06; 1.2120000e-06; 1.2130000e-06; 1.2140000e-06; 1.2150000e-06; 1.2160000e-06; 1.2170000e-06; 1.2180000e-06; 1.2190000e-06; 1.2200000e-06; 1.2210000e-06; 1.2220000e-06; 1.2230000e-06; 1.2240000e-06; 1.2250000e-06; 1.2260000e-06; 1.2270000e-06; 1.2280000e-06; 1.2290000e-06; 1.2300000e-06; 1.2310000e-06; 1.2320000e-06; 1.2330000e-06; 1.2340000e-06; 1.2350000e-06; 1.2360000e-06; 1.2370000e-06; 1.2380000e-06; 1.2390000e-06; 1.2400000e-06; 1.2410000e-06; 1.2420000e-06; 1.2430000e-06; 1.2440000e-06; 1.2450000e-06; 1.2460000e-06; 1.2470000e-06; 1.2480000e-06; 1.2490000e-06; 1.2500000e-06; 1.2510000e-06; 1.2520000e-06; 1.2530000e-06; 1.2540000e-06; 1.2550000e-06; 1.2560000e-06; 1.2570000e-06; 1.2580000e-06; 1.2590000e-06; 1.2600000e-06; 1.2610000e-06; 1.2620000e-06; 1.2630000e-06; 1.2640000e-06; 1.2650000e-06; 1.2660000e-06; 1.2670000e-06; 1.2680000e-06; 1.2690000e-06; 1.2700000e-06; 1.2710000e-06; 1.2720000e-06; 1.2730000e-06; 1.2740000e-06; 1.2750000e-06; 1.2760000e-06; 1.2770000e-06; 1.2780000e-06; 1.2790000e-06; 1.2800000e-06; 1.2810000e-06; 1.2820000e-06; 1.2830000e-06; 1.2840000e-06; 1.2850000e-06; 1.2860000e-06; 1.2870000e-06; 1.2880000e-06; 1.2890000e-06; 1.2900000e-06; 1.2910000e-06; 1.2920000e-06; 1.2930000e-06; 1.2940000e-06; 1.2950000e-06; 1.2960000e-06; 1.2970000e-06; 1.2980000e-06; 1.2990000e-06; 1.3000000e-06; 1.3010000e-06; 1.3020000e-06; 1.3030000e-06; 1.3040000e-06; 1.3050000e-06; 1.3060000e-06; 1.3070000e-06; 1.3080000e-06; 1.3090000e-06; 1.3100000e-06; 1.3110000e-06; 1.3120000e-06; 1.3130000e-06; 1.3140000e-06; 1.3150000e-06; 1.3160000e-06; 1.3170000e-06; 1.3180000e-06; 1.3190000e-06; 1.3200000e-06; 1.3210000e-06; 1.3220000e-06; 1.3230000e-06; 1.3240000e-06; 1.3250000e-06; 1.3260000e-06; 1.3270000e-06; 1.3280000e-06; 1.3290000e-06; 1.3300000e-06; 1.3310000e-06; 1.3320000e-06; 1.3330000e-06; 1.3340000e-06; 1.3350000e-06; 1.3360000e-06; 1.3370000e-06; 1.3380000e-06; 1.3390000e-06; 1.3400000e-06; 1.3410000e-06; 1.3420000e-06; 1.3430000e-06; 1.3440000e-06; 1.3450000e-06; 1.3460000e-06; 1.3470000e-06; 1.3480000e-06; 1.3490000e-06; 1.3500000e-06; 1.3510000e-06; 1.3520000e-06; 1.3530000e-06; 1.3540000e-06; 1.3550000e-06; 1.3560000e-06; 1.3570000e-06; 1.3580000e-06; 1.3590000e-06; 1.3600000e-06; 1.3610000e-06; 1.3620000e-06; 1.3630000e-06; 1.3640000e-06; 1.3650000e-06; 1.3660000e-06; 1.3670000e-06; 1.3680000e-06; 1.3690000e-06; 1.3700000e-06; 1.3710000e-06; 1.3720000e-06; 1.3730000e-06; 1.3740000e-06; 1.3750000e-06; 1.3760000e-06; 1.3770000e-06; 1.3780000e-06; 1.3790000e-06; 1.3800000e-06; 1.3810000e-06; 1.3820000e-06; 1.3830000e-06; 1.3840000e-06; 1.3850000e-06; 1.3860000e-06; 1.3870000e-06; 1.3880000e-06; 1.3890000e-06; 1.3900000e-06; 1.3910000e-06; 1.3920000e-06; 1.3930000e-06; 1.3940000e-06; 1.3950000e-06; 1.3960000e-06; 1.3970000e-06; 1.3980000e-06; 1.3990000e-06; 1.4000000e-06; 1.4010000e-06; 1.4020000e-06; 1.4030000e-06; 1.4040000e-06; 1.4050000e-06; 1.4060000e-06; 1.4070000e-06; 1.4080000e-06; 1.4090000e-06; 1.4100000e-06; 1.4110000e-06; 1.4120000e-06; 1.4130000e-06; 1.4140000e-06; 1.4150000e-06; 1.4160000e-06; 1.4170000e-06; 1.4180000e-06; 1.4190000e-06; 1.4200000e-06; 1.4210000e-06; 1.4220000e-06; 1.4230000e-06; 1.4240000e-06; 1.4250000e-06; 1.4260000e-06; 1.4270000e-06; 1.4280000e-06; 1.4290000e-06; 1.4300000e-06; 1.4310000e-06; 1.4320000e-06; 1.4330000e-06; 1.4340000e-06; 1.4350000e-06; 1.4360000e-06; 1.4370000e-06; 1.4380000e-06; 1.4390000e-06; 1.4400000e-06; 1.4410000e-06; 1.4420000e-06; 1.4430000e-06; 1.4440000e-06; 1.4450000e-06; 1.4460000e-06; 1.4470000e-06; 1.4480000e-06; 1.4490000e-06; 1.4500000e-06; 1.4510000e-06; 1.4520000e-06; 1.4530000e-06; 1.4540000e-06; 1.4550000e-06; 1.4560000e-06; 1.4570000e-06; 1.4580000e-06; 1.4590000e-06; 1.4600000e-06; 1.4610000e-06; 1.4620000e-06; 1.4630000e-06; 1.4640000e-06; 1.4650000e-06; 1.4660000e-06; 1.4670000e-06; 1.4680000e-06; 1.4690000e-06; 1.4700000e-06; 1.4710000e-06; 1.4720000e-06; 1.4730000e-06; 1.4740000e-06; 1.4750000e-06; 1.4760000e-06; 1.4770000e-06; 1.4780000e-06; 1.4790000e-06; 1.4800000e-06; 1.4810000e-06; 1.4820000e-06; 1.4830000e-06; 1.4840000e-06; 1.4850000e-06; 1.4860000e-06; 1.4870000e-06; 1.4880000e-06; 1.4890000e-06; 1.4900000e-06; 1.4910000e-06; 1.4920000e-06; 1.4930000e-06; 1.4940000e-06; 1.4950000e-06; 1.4960000e-06; 1.4970000e-06; 1.4980000e-06; 1.4990000e-06; 1.5000000e-06; 1.5010000e-06; 1.5020000e-06; 1.5030000e-06; 1.5040000e-06; 1.5050000e-06; 1.5060000e-06; 1.5070000e-06; 1.5080000e-06; 1.5090000e-06; 1.5100000e-06; 1.5110000e-06; 1.5120000e-06; 1.5130000e-06; 1.5140000e-06; 1.5150000e-06; 1.5160000e-06; 1.5170000e-06; 1.5180000e-06; 1.5190000e-06; 1.5200000e-06; 1.5210000e-06; 1.5220000e-06; 1.5230000e-06; 1.5240000e-06; 1.5250000e-06; 1.5260000e-06; 1.5270000e-06; 1.5280000e-06; 1.5290000e-06; 1.5300000e-06; 1.5310000e-06; 1.5320000e-06; 1.5330000e-06; 1.5340000e-06; 1.5350000e-06; 1.5360000e-06; 1.5370000e-06; 1.5380000e-06; 1.5390000e-06; 1.5400000e-06; 1.5410000e-06; 1.5420000e-06; 1.5430000e-06; 1.5440000e-06; 1.5450000e-06; 1.5460000e-06; 1.5470000e-06; 1.5480000e-06; 1.5490000e-06; 1.5500000e-06; 1.5510000e-06; 1.5520000e-06; 1.5530000e-06; 1.5540000e-06; 1.5550000e-06; 1.5560000e-06; 1.5570000e-06; 1.5580000e-06; 1.5590000e-06; 1.5600000e-06; 1.5610000e-06; 1.5620000e-06; 1.5630000e-06; 1.5640000e-06; 1.5650000e-06; 1.5660000e-06; 1.5670000e-06; 1.5680000e-06; 1.5690000e-06; 1.5700000e-06; 1.5710000e-06; 1.5720000e-06; 1.5730000e-06; 1.5740000e-06; 1.5750000e-06; 1.5760000e-06; 1.5770000e-06; 1.5780000e-06; 1.5790000e-06; 1.5800000e-06; 1.5810000e-06; 1.5820000e-06; 1.5830000e-06; 1.5840000e-06; 1.5850000e-06; 1.5860000e-06; 1.5870000e-06; 1.5880000e-06; 1.5890000e-06; 1.5900000e-06; 1.5910000e-06; 1.5920000e-06; 1.5930000e-06; 1.5940000e-06; 1.5950000e-06; 1.5960000e-06; 1.5970000e-06; 1.5980000e-06; 1.5990000e-06; 1.6000000e-06; 1.6010000e-06; 1.6020000e-06; 1.6030000e-06; 1.6040000e-06; 1.6050000e-06; 1.6060000e-06; 1.6070000e-06; 1.6080000e-06; 1.6090000e-06; 1.6100000e-06; 1.6110000e-06; 1.6120000e-06; 1.6130000e-06; 1.6140000e-06; 1.6150000e-06; 1.6160000e-06; 1.6170000e-06; 1.6180000e-06; 1.6190000e-06; 1.6200000e-06; 1.6210000e-06; 1.6220000e-06; 1.6230000e-06; 1.6240000e-06; 1.6250000e-06; 1.6260000e-06; 1.6270000e-06; 1.6280000e-06; 1.6290000e-06; 1.6300000e-06; 1.6310000e-06; 1.6320000e-06; 1.6330000e-06; 1.6340000e-06; 1.6350000e-06; 1.6360000e-06; 1.6370000e-06; 1.6380000e-06; 1.6390000e-06; 1.6400000e-06; 1.6410000e-06; 1.6420000e-06; 1.6430000e-06; 1.6440000e-06; 1.6450000e-06; 1.6460000e-06; 1.6470000e-06; 1.6480000e-06; 1.6490000e-06; 1.6500000e-06; 1.6510000e-06; 1.6520000e-06; 1.6530000e-06; 1.6540000e-06; 1.6550000e-06; 1.6560000e-06; 1.6570000e-06; 1.6580000e-06; 1.6590000e-06; 1.6600000e-06; 1.6610000e-06; 1.6620000e-06; 1.6630000e-06; 1.6640000e-06; 1.6650000e-06; 1.6660000e-06; 1.6670000e-06; 1.6680000e-06; 1.6690000e-06; 1.6700000e-06; 1.6710000e-06; 1.6720000e-06; 1.6730000e-06; 1.6740000e-06; 1.6750000e-06; 1.6760000e-06; 1.6770000e-06; 1.6780000e-06; 1.6790000e-06; 1.6800000e-06; 1.6810000e-06; 1.6820000e-06; 1.6830000e-06; 1.6840000e-06; 1.6850000e-06; 1.6860000e-06; 1.6870000e-06; 1.6880000e-06; 1.6890000e-06; 1.6900000e-06; 1.6910000e-06; 1.6920000e-06; 1.6930000e-06; 1.6940000e-06; 1.6950000e-06; 1.6960000e-06; 1.6970000e-06; 1.6980000e-06; 1.6990000e-06; 1.7000000e-06; 1.7010000e-06; 1.7020000e-06; 1.7030000e-06; 1.7040000e-06; 1.7050000e-06; 1.7060000e-06; 1.7070000e-06; 1.7080000e-06; 1.7090000e-06; 1.7100000e-06; 1.7110000e-06; 1.7120000e-06; 1.7130000e-06; 1.7140000e-06; 1.7150000e-06; 1.7160000e-06; 1.7170000e-06; 1.7180000e-06; 1.7190000e-06; 1.7200000e-06; 1.7210000e-06; 1.7220000e-06; 1.7230000e-06; 1.7240000e-06; 1.7250000e-06; 1.7260000e-06; 1.7270000e-06; 1.7280000e-06; 1.7290000e-06; 1.7300000e-06; 1.7310000e-06; 1.7320000e-06; 1.7330000e-06; 1.7340000e-06; 1.7350000e-06; 1.7360000e-06; 1.7370000e-06; 1.7380000e-06; 1.7390000e-06; 1.7400000e-06; 1.7410000e-06; 1.7420000e-06; 1.7430000e-06; 1.7440000e-06; 1.7450000e-06; 1.7460000e-06; 1.7470000e-06; 1.7480000e-06; 1.7490000e-06; 1.7500000e-06; 1.7510000e-06; 1.7520000e-06; 1.7530000e-06; 1.7540000e-06; 1.7550000e-06; 1.7560000e-06; 1.7570000e-06; 1.7580000e-06; 1.7590000e-06; 1.7600000e-06; 1.7610000e-06; 1.7620000e-06; 1.7630000e-06; 1.7640000e-06; 1.7650000e-06; 1.7660000e-06; 1.7670000e-06; 1.7680000e-06; 1.7690000e-06; 1.7700000e-06; 1.7710000e-06; 1.7720000e-06; 1.7730000e-06; 1.7740000e-06; 1.7750000e-06; 1.7760000e-06; 1.7770000e-06; 1.7780000e-06; 1.7790000e-06; 1.7800000e-06; 1.7810000e-06; 1.7820000e-06; 1.7830000e-06; 1.7840000e-06; 1.7850000e-06; 1.7860000e-06; 1.7870000e-06; 1.7880000e-06; 1.7890000e-06; 1.7900000e-06; 1.7910000e-06; 1.7920000e-06; 1.7930000e-06; 1.7940000e-06; 1.7950000e-06; 1.7960000e-06; 1.7970000e-06; 1.7980000e-06; 1.7990000e-06; 1.8000000e-06; 1.8010000e-06; 1.8020000e-06; 1.8030000e-06; 1.8040000e-06; 1.8050000e-06; 1.8060000e-06; 1.8070000e-06; 1.8080000e-06; 1.8090000e-06; 1.8100000e-06; 1.8110000e-06; 1.8120000e-06; 1.8130000e-06; 1.8140000e-06; 1.8150000e-06; 1.8160000e-06; 1.8170000e-06; 1.8180000e-06; 1.8190000e-06; 1.8200000e-06; 1.8210000e-06; 1.8220000e-06; 1.8230000e-06; 1.8240000e-06; 1.8250000e-06; 1.8260000e-06; 1.8270000e-06; 1.8280000e-06; 1.8290000e-06; 1.8300000e-06; 1.8310000e-06; 1.8320000e-06; 1.8330000e-06; 1.8340000e-06; 1.8350000e-06; 1.8360000e-06; 1.8370000e-06; 1.8380000e-06; 1.8390000e-06; 1.8400000e-06; 1.8410000e-06; 1.8420000e-06; 1.8430000e-06; 1.8440000e-06; 1.8450000e-06; 1.8460000e-06; 1.8470000e-06; 1.8480000e-06; 1.8490000e-06; 1.8500000e-06; 1.8510000e-06; 1.8520000e-06; 1.8530000e-06; 1.8540000e-06; 1.8550000e-06; 1.8560000e-06; 1.8570000e-06; 1.8580000e-06; 1.8590000e-06; 1.8600000e-06; 1.8610000e-06; 1.8620000e-06; 1.8630000e-06; 1.8640000e-06; 1.8650000e-06; 1.8660000e-06; 1.8670000e-06; 1.8680000e-06; 1.8690000e-06; 1.8700000e-06; 1.8710000e-06; 1.8720000e-06; 1.8730000e-06; 1.8740000e-06; 1.8750000e-06; 1.8760000e-06; 1.8770000e-06; 1.8780000e-06; 1.8790000e-06; 1.8800000e-06; 1.8810000e-06; 1.8820000e-06; 1.8830000e-06; 1.8840000e-06; 1.8850000e-06; 1.8860000e-06; 1.8870000e-06; 1.8880000e-06; 1.8890000e-06; 1.8900000e-06; 1.8910000e-06; 1.8920000e-06; 1.8930000e-06; 1.8940000e-06; 1.8950000e-06; 1.8960000e-06; 1.8970000e-06; 1.8980000e-06; 1.8990000e-06; 1.9000000e-06; 1.9010000e-06; 1.9020000e-06; 1.9030000e-06; 1.9040000e-06; 1.9050000e-06; 1.9060000e-06; 1.9070000e-06; 1.9080000e-06; 1.9090000e-06; 1.9100000e-06; 1.9110000e-06; 1.9120000e-06; 1.9130000e-06; 1.9140000e-06; 1.9150000e-06; 1.9160000e-06; 1.9170000e-06; 1.9180000e-06; 1.9190000e-06; 1.9200000e-06; 1.9210000e-06; 1.9220000e-06; 1.9230000e-06; 1.9240000e-06; 1.9250000e-06; 1.9260000e-06; 1.9270000e-06; 1.9280000e-06; 1.9290000e-06; 1.9300000e-06; 1.9310000e-06; 1.9320000e-06; 1.9330000e-06; 1.9340000e-06; 1.9350000e-06; 1.9360000e-06; 1.9370000e-06; 1.9380000e-06; 1.9390000e-06; 1.9400000e-06; 1.9410000e-06; 1.9420000e-06; 1.9430000e-06; 1.9440000e-06; 1.9450000e-06; 1.9460000e-06; 1.9470000e-06; 1.9480000e-06; 1.9490000e-06; 1.9500000e-06; 1.9510000e-06; 1.9520000e-06; 1.9530000e-06; 1.9540000e-06; 1.9550000e-06; 1.9560000e-06; 1.9570000e-06; 1.9580000e-06; 1.9590000e-06; 1.9600000e-06; 1.9610000e-06; 1.9620000e-06; 1.9630000e-06; 1.9640000e-06; 1.9650000e-06; 1.9660000e-06; 1.9670000e-06; 1.9680000e-06; 1.9690000e-06; 1.9700000e-06; 1.9710000e-06; 1.9720000e-06; 1.9730000e-06; 1.9740000e-06; 1.9750000e-06; 1.9760000e-06; 1.9770000e-06; 1.9780000e-06; 1.9790000e-06; 1.9800000e-06; 1.9810000e-06; 1.9820000e-06; 1.9830000e-06; 1.9840000e-06; 1.9850000e-06; 1.9860000e-06; 1.9870000e-06; 1.9880000e-06; 1.9890000e-06; 1.9900000e-06; 1.9910000e-06; 1.9920000e-06; 1.9930000e-06; 1.9940000e-06; 1.9950000e-06; 1.9960000e-06; 1.9970000e-06; 1.9980000e-06; 1.9990000e-06; 2.0000000e-06; 2.0010000e-06; 2.0020000e-06; 2.0030000e-06; 2.0040000e-06; 2.0050000e-06; 2.0060000e-06; 2.0070000e-06; 2.0080000e-06; 2.0090000e-06; 2.0100000e-06; 2.0110000e-06; 2.0120000e-06; 2.0130000e-06; 2.0140000e-06; 2.0150000e-06; 2.0160000e-06; 2.0170000e-06; 2.0180000e-06; 2.0190000e-06; 2.0200000e-06; 2.0210000e-06; 2.0220000e-06; 2.0230000e-06; 2.0240000e-06; 2.0250000e-06; 2.0260000e-06; 2.0270000e-06; 2.0280000e-06; 2.0290000e-06; 2.0300000e-06; 2.0310000e-06; 2.0320000e-06; 2.0330000e-06; 2.0340000e-06; 2.0350000e-06; 2.0360000e-06; 2.0370000e-06; 2.0380000e-06; 2.0390000e-06; 2.0400000e-06; 2.0410000e-06; 2.0420000e-06; 2.0430000e-06; 2.0440000e-06; 2.0450000e-06; 2.0460000e-06; 2.0470000e-06; 2.0480000e-06; 2.0490000e-06; 2.0500000e-06; 2.0510000e-06; 2.0520000e-06; 2.0530000e-06; 2.0540000e-06; 2.0550000e-06; 2.0560000e-06; 2.0570000e-06; 2.0580000e-06; 2.0590000e-06; 2.0600000e-06; 2.0610000e-06; 2.0620000e-06; 2.0630000e-06; 2.0640000e-06; 2.0650000e-06; 2.0660000e-06; 2.0670000e-06; 2.0680000e-06; 2.0690000e-06; 2.0700000e-06; 2.0710000e-06; 2.0720000e-06; 2.0730000e-06; 2.0740000e-06; 2.0750000e-06; 2.0760000e-06; 2.0770000e-06; 2.0780000e-06; 2.0790000e-06; 2.0800000e-06; 2.0810000e-06; 2.0820000e-06; 2.0830000e-06; 2.0840000e-06; 2.0850000e-06; 2.0860000e-06; 2.0870000e-06; 2.0880000e-06; 2.0890000e-06; 2.0900000e-06; 2.0910000e-06; 2.0920000e-06; 2.0930000e-06; 2.0940000e-06; 2.0950000e-06; 2.0960000e-06; 2.0970000e-06; 2.0980000e-06; 2.0990000e-06; 2.1000000e-06; 2.1010000e-06; 2.1020000e-06; 2.1030000e-06; 2.1040000e-06; 2.1050000e-06; 2.1060000e-06; 2.1070000e-06; 2.1080000e-06; 2.1090000e-06; 2.1100000e-06; 2.1110000e-06; 2.1120000e-06; 2.1130000e-06; 2.1140000e-06; 2.1150000e-06; 2.1160000e-06; 2.1170000e-06; 2.1180000e-06; 2.1190000e-06; 2.1200000e-06; 2.1210000e-06; 2.1220000e-06; 2.1230000e-06; 2.1240000e-06; 2.1250000e-06; 2.1260000e-06; 2.1270000e-06; 2.1280000e-06; 2.1290000e-06; 2.1300000e-06; 2.1310000e-06; 2.1320000e-06; 2.1330000e-06; 2.1340000e-06; 2.1350000e-06; 2.1360000e-06; 2.1370000e-06; 2.1380000e-06; 2.1390000e-06; 2.1400000e-06; 2.1410000e-06; 2.1420000e-06; 2.1430000e-06; 2.1440000e-06; 2.1450000e-06; 2.1460000e-06; 2.1470000e-06; 2.1480000e-06; 2.1490000e-06; 2.1500000e-06; 2.1510000e-06; 2.1520000e-06; 2.1530000e-06; 2.1540000e-06; 2.1550000e-06; 2.1560000e-06; 2.1570000e-06; 2.1580000e-06; 2.1590000e-06; 2.1600000e-06; 2.1610000e-06; 2.1620000e-06; 2.1630000e-06; 2.1640000e-06; 2.1650000e-06; 2.1660000e-06; 2.1670000e-06; 2.1680000e-06; 2.1690000e-06; 2.1700000e-06; 2.1710000e-06; 2.1720000e-06; 2.1730000e-06; 2.1740000e-06; 2.1750000e-06; 2.1760000e-06; 2.1770000e-06; 2.1780000e-06; 2.1790000e-06; 2.1800000e-06; 2.1810000e-06; 2.1820000e-06; 2.1830000e-06; 2.1840000e-06; 2.1850000e-06; 2.1860000e-06; 2.1870000e-06; 2.1880000e-06; 2.1890000e-06; 2.1900000e-06; 2.1910000e-06; 2.1920000e-06; 2.1930000e-06; 2.1940000e-06; 2.1950000e-06; 2.1960000e-06; 2.1970000e-06; 2.1980000e-06; 2.1990000e-06; 2.2000000e-06; 2.2010000e-06; 2.2020000e-06; 2.2030000e-06; 2.2040000e-06; 2.2050000e-06; 2.2060000e-06; 2.2070000e-06; 2.2080000e-06; 2.2090000e-06; 2.2100000e-06; 2.2110000e-06; 2.2120000e-06; 2.2130000e-06; 2.2140000e-06; 2.2150000e-06; 2.2160000e-06; 2.2170000e-06; 2.2180000e-06; 2.2190000e-06; 2.2200000e-06; 2.2210000e-06; 2.2220000e-06; 2.2230000e-06; 2.2240000e-06; 2.2250000e-06; 2.2260000e-06; 2.2270000e-06; 2.2280000e-06; 2.2290000e-06; 2.2300000e-06; 2.2310000e-06; 2.2320000e-06; 2.2330000e-06; 2.2340000e-06; 2.2350000e-06; 2.2360000e-06; 2.2370000e-06; 2.2380000e-06; 2.2390000e-06; 2.2400000e-06; 2.2410000e-06; 2.2420000e-06; 2.2430000e-06; 2.2440000e-06; 2.2450000e-06; 2.2460000e-06; 2.2470000e-06; 2.2480000e-06; 2.2490000e-06; 2.2500000e-06; 2.2510000e-06; 2.2520000e-06; 2.2530000e-06; 2.2540000e-06; 2.2550000e-06; 2.2560000e-06; 2.2570000e-06; 2.2580000e-06; 2.2590000e-06; 2.2600000e-06; 2.2610000e-06; 2.2620000e-06; 2.2630000e-06; 2.2640000e-06; 2.2650000e-06; 2.2660000e-06; 2.2670000e-06; 2.2680000e-06; 2.2690000e-06; 2.2700000e-06; 2.2710000e-06; 2.2720000e-06; 2.2730000e-06; 2.2740000e-06; 2.2750000e-06; 2.2760000e-06; 2.2770000e-06; 2.2780000e-06; 2.2790000e-06; 2.2800000e-06; 2.2810000e-06; 2.2820000e-06; 2.2830000e-06; 2.2840000e-06; 2.2850000e-06; 2.2860000e-06; 2.2870000e-06; 2.2880000e-06; 2.2890000e-06; 2.2900000e-06; 2.2910000e-06; 2.2920000e-06; 2.2930000e-06; 2.2940000e-06; 2.2950000e-06; 2.2960000e-06; 2.2970000e-06; 2.2980000e-06; 2.2990000e-06; 2.3000000e-06; 2.3010000e-06; 2.3020000e-06; 2.3030000e-06; 2.3040000e-06; 2.3050000e-06; 2.3060000e-06; 2.3070000e-06; 2.3080000e-06; 2.3090000e-06; 2.3100000e-06; 2.3110000e-06; 2.3120000e-06; 2.3130000e-06; 2.3140000e-06; 2.3150000e-06; 2.3160000e-06; 2.3170000e-06; 2.3180000e-06; 2.3190000e-06; 2.3200000e-06; 2.3210000e-06; 2.3220000e-06; 2.3230000e-06; 2.3240000e-06; 2.3250000e-06; 2.3260000e-06; 2.3270000e-06; 2.3280000e-06; 2.3290000e-06; 2.3300000e-06; 2.3310000e-06; 2.3320000e-06; 2.3330000e-06; 2.3340000e-06; 2.3350000e-06; 2.3360000e-06; 2.3370000e-06; 2.3380000e-06; 2.3390000e-06; 2.3400000e-06; 2.3410000e-06; 2.3420000e-06; 2.3430000e-06; 2.3440000e-06; 2.3450000e-06; 2.3460000e-06; 2.3470000e-06; 2.3480000e-06; 2.3490000e-06; 2.3500000e-06; 2.3510000e-06; 2.3520000e-06; 2.3530000e-06; 2.3540000e-06; 2.3550000e-06; 2.3560000e-06; 2.3570000e-06; 2.3580000e-06; 2.3590000e-06; 2.3600000e-06; 2.3610000e-06; 2.3620000e-06; 2.3630000e-06; 2.3640000e-06; 2.3650000e-06; 2.3660000e-06; 2.3670000e-06; 2.3680000e-06; 2.3690000e-06; 2.3700000e-06; 2.3710000e-06; 2.3720000e-06; 2.3730000e-06; 2.3740000e-06; 2.3750000e-06; 2.3760000e-06; 2.3770000e-06; 2.3780000e-06; 2.3790000e-06; 2.3800000e-06; 2.3810000e-06; 2.3820000e-06; 2.3830000e-06; 2.3840000e-06; 2.3850000e-06; 2.3860000e-06; 2.3870000e-06; 2.3880000e-06; 2.3890000e-06; 2.3900000e-06; 2.3910000e-06; 2.3920000e-06; 2.3930000e-06; 2.3940000e-06; 2.3950000e-06; 2.3960000e-06; 2.3970000e-06; 2.3980000e-06; 2.3990000e-06; 2.4000000e-06; 2.4010000e-06; 2.4020000e-06; 2.4030000e-06; 2.4040000e-06; 2.4050000e-06; 2.4060000e-06; 2.4070000e-06; 2.4080000e-06; 2.4090000e-06; 2.4100000e-06; 2.4110000e-06; 2.4120000e-06; 2.4130000e-06; 2.4140000e-06; 2.4150000e-06; 2.4160000e-06; 2.4170000e-06; 2.4180000e-06; 2.4190000e-06; 2.4200000e-06; 2.4210000e-06; 2.4220000e-06; 2.4230000e-06; 2.4240000e-06; 2.4250000e-06; 2.4260000e-06; 2.4270000e-06; 2.4280000e-06; 2.4290000e-06; 2.4300000e-06; 2.4310000e-06; 2.4320000e-06; 2.4330000e-06; 2.4340000e-06; 2.4350000e-06; 2.4360000e-06; 2.4370000e-06; 2.4380000e-06; 2.4390000e-06; 2.4400000e-06; 2.4410000e-06; 2.4420000e-06; 2.4430000e-06; 2.4440000e-06; 2.4450000e-06; 2.4460000e-06; 2.4470000e-06; 2.4480000e-06; 2.4490000e-06; 2.4500000e-06; 2.4510000e-06; 2.4520000e-06; 2.4530000e-06; 2.4540000e-06; 2.4550000e-06; 2.4560000e-06; 2.4570000e-06; 2.4580000e-06; 2.4590000e-06; 2.4600000e-06; 2.4610000e-06; 2.4620000e-06; 2.4630000e-06; 2.4640000e-06; 2.4650000e-06; 2.4660000e-06; 2.4670000e-06; 2.4680000e-06; 2.4690000e-06; 2.4700000e-06; 2.4710000e-06; 2.4720000e-06; 2.4730000e-06; 2.4740000e-06; 2.4750000e-06; 2.4760000e-06; 2.4770000e-06; 2.4780000e-06; 2.4790000e-06; 2.4800000e-06; 2.4810000e-06; 2.4820000e-06; 2.4830000e-06; 2.4840000e-06; 2.4850000e-06; 2.4860000e-06; 2.4870000e-06; 2.4880000e-06; 2.4890000e-06; 2.4900000e-06; 2.4910000e-06; 2.4920000e-06; 2.4930000e-06; 2.4940000e-06; 2.4950000e-06; 2.4960000e-06; 2.4970000e-06; 2.4980000e-06; 2.4990000e-06; 2.5000000e-06; 2.5010000e-06; 2.5020000e-06; 2.5030000e-06; 2.5040000e-06; 2.5050000e-06; 2.5060000e-06; 2.5070000e-06; 2.5080000e-06; 2.5090000e-06; 2.5100000e-06; 2.5110000e-06; 2.5120000e-06; 2.5130000e-06; 2.5140000e-06; 2.5150000e-06; 2.5160000e-06; 2.5170000e-06; 2.5180000e-06; 2.5190000e-06; 2.5200000e-06; 2.5210000e-06; 2.5220000e-06; 2.5230000e-06; 2.5240000e-06; 2.5250000e-06; 2.5260000e-06; 2.5270000e-06; 2.5280000e-06; 2.5290000e-06; 2.5300000e-06; 2.5310000e-06; 2.5320000e-06; 2.5330000e-06; 2.5340000e-06; 2.5350000e-06; 2.5360000e-06; 2.5370000e-06; 2.5380000e-06; 2.5390000e-06; 2.5400000e-06; 2.5410000e-06; 2.5420000e-06; 2.5430000e-06; 2.5440000e-06; 2.5450000e-06; 2.5460000e-06; 2.5470000e-06; 2.5480000e-06; 2.5490000e-06; 2.5500000e-06; 2.5510000e-06; 2.5520000e-06; 2.5530000e-06; 2.5540000e-06; 2.5550000e-06; 2.5560000e-06; 2.5570000e-06; 2.5580000e-06; 2.5590000e-06; 2.5600000e-06; 2.5610000e-06; 2.5620000e-06; 2.5630000e-06; 2.5640000e-06; 2.5650000e-06; 2.5660000e-06; 2.5670000e-06; 2.5680000e-06; 2.5690000e-06; 2.5700000e-06; 2.5710000e-06; 2.5720000e-06; 2.5730000e-06; 2.5740000e-06; 2.5750000e-06; 2.5760000e-06; 2.5770000e-06; 2.5780000e-06; 2.5790000e-06; 2.5800000e-06; 2.5810000e-06; 2.5820000e-06; 2.5830000e-06; 2.5840000e-06; 2.5850000e-06; 2.5860000e-06; 2.5870000e-06; 2.5880000e-06; 2.5890000e-06; 2.5900000e-06; 2.5910000e-06; 2.5920000e-06; 2.5930000e-06; 2.5940000e-06; 2.5950000e-06; 2.5960000e-06; 2.5970000e-06; 2.5980000e-06; 2.5990000e-06; 2.6000000e-06; 2.6010000e-06; 2.6020000e-06; 2.6030000e-06; 2.6040000e-06; 2.6050000e-06; 2.6060000e-06; 2.6070000e-06; 2.6080000e-06; 2.6090000e-06; 2.6100000e-06; 2.6110000e-06; 2.6120000e-06; 2.6130000e-06; 2.6140000e-06; 2.6150000e-06; 2.6160000e-06; 2.6170000e-06; 2.6180000e-06; 2.6190000e-06; 2.6200000e-06; 2.6210000e-06; 2.6220000e-06; 2.6230000e-06; 2.6240000e-06; 2.6250000e-06; 2.6260000e-06; 2.6270000e-06; 2.6280000e-06; 2.6290000e-06; 2.6300000e-06; 2.6310000e-06; 2.6320000e-06; 2.6330000e-06; 2.6340000e-06; 2.6350000e-06; 2.6360000e-06; 2.6370000e-06; 2.6380000e-06; 2.6390000e-06; 2.6400000e-06; 2.6410000e-06; 2.6420000e-06; 2.6430000e-06; 2.6440000e-06; 2.6450000e-06; 2.6460000e-06; 2.6470000e-06; 2.6480000e-06; 2.6490000e-06; 2.6500000e-06; 2.6510000e-06; 2.6520000e-06; 2.6530000e-06; 2.6540000e-06; 2.6550000e-06; 2.6560000e-06; 2.6570000e-06; 2.6580000e-06; 2.6590000e-06; 2.6600000e-06; 2.6610000e-06; 2.6620000e-06; 2.6630000e-06; 2.6640000e-06; 2.6650000e-06; 2.6660000e-06; 2.6670000e-06; 2.6680000e-06; 2.6690000e-06; 2.6700000e-06; 2.6710000e-06; 2.6720000e-06; 2.6730000e-06; 2.6740000e-06; 2.6750000e-06; 2.6760000e-06; 2.6770000e-06; 2.6780000e-06; 2.6790000e-06; 2.6800000e-06; 2.6810000e-06; 2.6820000e-06; 2.6830000e-06; 2.6840000e-06; 2.6850000e-06; 2.6860000e-06; 2.6870000e-06; 2.6880000e-06; 2.6890000e-06; 2.6900000e-06; 2.6910000e-06; 2.6920000e-06; 2.6930000e-06; 2.6940000e-06; 2.6950000e-06; 2.6960000e-06; 2.6970000e-06; 2.6980000e-06; 2.6990000e-06; 2.7000000e-06; 2.7010000e-06; 2.7020000e-06; 2.7030000e-06; 2.7040000e-06; 2.7050000e-06; 2.7060000e-06; 2.7070000e-06; 2.7080000e-06; 2.7090000e-06; 2.7100000e-06; 2.7110000e-06; 2.7120000e-06; 2.7130000e-06; 2.7140000e-06; 2.7150000e-06; 2.7160000e-06; 2.7170000e-06; 2.7180000e-06; 2.7190000e-06; 2.7200000e-06; 2.7210000e-06; 2.7220000e-06; 2.7230000e-06; 2.7240000e-06; 2.7250000e-06; 2.7260000e-06; 2.7270000e-06; 2.7280000e-06; 2.7290000e-06; 2.7300000e-06; 2.7310000e-06; 2.7320000e-06; 2.7330000e-06; 2.7340000e-06; 2.7350000e-06; 2.7360000e-06; 2.7370000e-06; 2.7380000e-06; 2.7390000e-06; 2.7400000e-06; 2.7410000e-06; 2.7420000e-06; 2.7430000e-06; 2.7440000e-06; 2.7450000e-06; 2.7460000e-06; 2.7470000e-06; 2.7480000e-06; 2.7490000e-06; 2.7500000e-06; 2.7510000e-06; 2.7520000e-06; 2.7530000e-06; 2.7540000e-06; 2.7550000e-06; 2.7560000e-06; 2.7570000e-06; 2.7580000e-06; 2.7590000e-06; 2.7600000e-06; 2.7610000e-06; 2.7620000e-06; 2.7630000e-06; 2.7640000e-06; 2.7650000e-06; 2.7660000e-06; 2.7670000e-06; 2.7680000e-06; 2.7690000e-06; 2.7700000e-06; 2.7710000e-06; 2.7720000e-06; 2.7730000e-06; 2.7740000e-06; 2.7750000e-06; 2.7760000e-06; 2.7770000e-06; 2.7780000e-06; 2.7790000e-06; 2.7800000e-06; 2.7810000e-06; 2.7820000e-06; 2.7830000e-06; 2.7840000e-06; 2.7850000e-06; 2.7860000e-06; 2.7870000e-06; 2.7880000e-06; 2.7890000e-06; 2.7900000e-06; 2.7910000e-06; 2.7920000e-06; 2.7930000e-06; 2.7940000e-06; 2.7950000e-06; 2.7960000e-06; 2.7970000e-06; 2.7980000e-06; 2.7990000e-06; 2.8000000e-06; 2.8010000e-06; 2.8020000e-06; 2.8030000e-06; 2.8040000e-06; 2.8050000e-06; 2.8060000e-06; 2.8070000e-06; 2.8080000e-06; 2.8090000e-06; 2.8100000e-06; 2.8110000e-06; 2.8120000e-06; 2.8130000e-06; 2.8140000e-06; 2.8150000e-06; 2.8160000e-06; 2.8170000e-06; 2.8180000e-06; 2.8190000e-06; 2.8200000e-06; 2.8210000e-06; 2.8220000e-06; 2.8230000e-06; 2.8240000e-06; 2.8250000e-06; 2.8260000e-06; 2.8270000e-06; 2.8280000e-06; 2.8290000e-06; 2.8300000e-06; 2.8310000e-06; 2.8320000e-06; 2.8330000e-06; 2.8340000e-06; 2.8350000e-06; 2.8360000e-06; 2.8370000e-06; 2.8380000e-06; 2.8390000e-06; 2.8400000e-06; 2.8410000e-06; 2.8420000e-06; 2.8430000e-06; 2.8440000e-06; 2.8450000e-06; 2.8460000e-06; 2.8470000e-06; 2.8480000e-06; 2.8490000e-06; 2.8500000e-06; 2.8510000e-06; 2.8520000e-06; 2.8530000e-06; 2.8540000e-06; 2.8550000e-06; 2.8560000e-06; 2.8570000e-06; 2.8580000e-06; 2.8590000e-06; 2.8600000e-06; 2.8610000e-06; 2.8620000e-06; 2.8630000e-06; 2.8640000e-06; 2.8650000e-06; 2.8660000e-06; 2.8670000e-06; 2.8680000e-06; 2.8690000e-06; 2.8700000e-06; 2.8710000e-06; 2.8720000e-06; 2.8730000e-06; 2.8740000e-06; 2.8750000e-06; 2.8760000e-06; 2.8770000e-06; 2.8780000e-06; 2.8790000e-06; 2.8800000e-06; 2.8810000e-06; 2.8820000e-06; 2.8830000e-06; 2.8840000e-06; 2.8850000e-06; 2.8860000e-06; 2.8870000e-06; 2.8880000e-06; 2.8890000e-06; 2.8900000e-06; 2.8910000e-06; 2.8920000e-06; 2.8930000e-06; 2.8940000e-06; 2.8950000e-06; 2.8960000e-06; 2.8970000e-06; 2.8980000e-06; 2.8990000e-06; 2.9000000e-06; 2.9010000e-06; 2.9020000e-06; 2.9030000e-06; 2.9040000e-06; 2.9050000e-06; 2.9060000e-06; 2.9070000e-06; 2.9080000e-06; 2.9090000e-06; 2.9100000e-06; 2.9110000e-06; 2.9120000e-06; 2.9130000e-06; 2.9140000e-06; 2.9150000e-06; 2.9160000e-06; 2.9170000e-06; 2.9180000e-06; 2.9190000e-06; 2.9200000e-06; 2.9210000e-06; 2.9220000e-06; 2.9230000e-06; 2.9240000e-06; 2.9250000e-06; 2.9260000e-06; 2.9270000e-06; 2.9280000e-06; 2.9290000e-06; 2.9300000e-06; 2.9310000e-06; 2.9320000e-06; 2.9330000e-06; 2.9340000e-06; 2.9350000e-06; 2.9360000e-06; 2.9370000e-06; 2.9380000e-06; 2.9390000e-06; 2.9400000e-06; 2.9410000e-06; 2.9420000e-06; 2.9430000e-06; 2.9440000e-06; 2.9450000e-06; 2.9460000e-06; 2.9470000e-06; 2.9480000e-06; 2.9490000e-06; 2.9500000e-06; 2.9510000e-06; 2.9520000e-06; 2.9530000e-06; 2.9540000e-06; 2.9550000e-06; 2.9560000e-06; 2.9570000e-06; 2.9580000e-06; 2.9590000e-06; 2.9600000e-06; 2.9610000e-06; 2.9620000e-06; 2.9630000e-06; 2.9640000e-06; 2.9650000e-06; 2.9660000e-06; 2.9670000e-06; 2.9680000e-06; 2.9690000e-06; 2.9700000e-06; 2.9710000e-06; 2.9720000e-06; 2.9730000e-06; 2.9740000e-06; 2.9750000e-06; 2.9760000e-06; 2.9770000e-06; 2.9780000e-06; 2.9790000e-06; 2.9800000e-06; 2.9810000e-06; 2.9820000e-06; 2.9830000e-06; 2.9840000e-06; 2.9850000e-06; 2.9860000e-06; 2.9870000e-06; 2.9880000e-06; 2.9890000e-06; 2.9900000e-06; 2.9910000e-06; 2.9920000e-06; 2.9930000e-06; 2.9940000e-06; 2.9950000e-06; 2.9960000e-06; 2.9970000e-06; 2.9980000e-06; 2.9990000e-06; 3.0000000e-06; 3.0010000e-06; 3.0020000e-06; 3.0030000e-06; 3.0040000e-06; 3.0050000e-06; 3.0060000e-06; 3.0070000e-06; 3.0080000e-06; 3.0090000e-06; 3.0100000e-06; 3.0110000e-06; 3.0120000e-06; 3.0130000e-06; 3.0140000e-06; 3.0150000e-06; 3.0160000e-06; 3.0170000e-06; 3.0180000e-06; 3.0190000e-06; 3.0200000e-06; 3.0210000e-06; 3.0220000e-06; 3.0230000e-06; 3.0240000e-06; 3.0250000e-06; 3.0260000e-06; 3.0270000e-06; 3.0280000e-06; 3.0290000e-06; 3.0300000e-06; 3.0310000e-06; 3.0320000e-06; 3.0330000e-06; 3.0340000e-06; 3.0350000e-06; 3.0360000e-06; 3.0370000e-06; 3.0380000e-06; 3.0390000e-06; 3.0400000e-06; 3.0410000e-06; 3.0420000e-06; 3.0430000e-06; 3.0440000e-06; 3.0450000e-06; 3.0460000e-06; 3.0470000e-06; 3.0480000e-06; 3.0490000e-06; 3.0500000e-06; 3.0510000e-06; 3.0520000e-06; 3.0530000e-06; 3.0540000e-06; 3.0550000e-06; 3.0560000e-06; 3.0570000e-06; 3.0580000e-06; 3.0590000e-06; 3.0600000e-06; 3.0610000e-06; 3.0620000e-06; 3.0630000e-06; 3.0640000e-06; 3.0650000e-06; 3.0660000e-06; 3.0670000e-06; 3.0680000e-06; 3.0690000e-06; 3.0700000e-06; 3.0710000e-06; 3.0720000e-06; 3.0730000e-06; 3.0740000e-06; 3.0750000e-06; 3.0760000e-06; 3.0770000e-06; 3.0780000e-06; 3.0790000e-06; 3.0800000e-06; 3.0810000e-06; 3.0820000e-06; 3.0830000e-06; 3.0840000e-06; 3.0850000e-06; 3.0860000e-06; 3.0870000e-06; 3.0880000e-06; 3.0890000e-06; 3.0900000e-06; 3.0910000e-06; 3.0920000e-06; 3.0930000e-06; 3.0940000e-06; 3.0950000e-06; 3.0960000e-06; 3.0970000e-06; 3.0980000e-06; 3.0990000e-06];
	kref = [Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; Inf; 1.0931559e-04; 1.0976360e-04; 1.1091928e-04; 1.1234491e-04; 1.1005010e-04; 1.0636399e-04; 1.0382261e-04; 1.0170746e-04; 1.0021744e-04; 9.8560863e-05; 9.7220118e-05; 9.5637512e-05; 9.3679233e-05; 9.1567259e-05; 8.9839712e-05; 8.7346318e-05; 8.4832179e-05; 8.2474338e-05; 8.0168395e-05; 7.7722289e-05; 7.5203132e-05; 7.2703651e-05; 7.0146043e-05; 6.7624654e-05; 6.5353014e-05; 6.3293015e-05; 6.1430824e-05; 5.9733052e-05; 5.8170959e-05; 5.6704216e-05; 5.5398954e-05; 5.4167079e-05; 5.3023594e-05; 5.1956505e-05; 5.0875214e-05; 4.9811027e-05; 4.8736424e-05; 4.7637026e-05; 4.6550403e-05; 4.5454628e-05; 4.4421285e-05; 4.3407955e-05; 4.2487212e-05; 4.1611825e-05; 4.0772198e-05; 3.9991531e-05; 3.9249925e-05; 3.8573055e-05; 3.7867350e-05; 3.7220297e-05; 3.6624829e-05; 3.6053502e-05; 3.5505064e-05; 3.4969859e-05; 3.4455686e-05; 3.3945306e-05; 3.3459779e-05; 3.2970955e-05; 3.2489323e-05; 3.2009620e-05; 3.1539245e-05; 3.1077960e-05; 3.0630840e-05; 3.0193539e-05; 2.9771743e-05; 2.9365362e-05; 2.8969045e-05; 2.8573374e-05; 2.8201602e-05; 2.7826876e-05; 2.7469281e-05; 2.7130863e-05; 2.6790088e-05; 2.6466061e-05; 2.6151394e-05; 2.5851716e-05; 2.5553572e-05; 2.5262699e-05; 2.4972178e-05; 2.4689247e-05; 2.4408147e-05; 2.4142109e-05; 2.3885323e-05; 2.3623979e-05; 2.3368135e-05; 2.3124974e-05; 2.2884052e-05; 2.2643587e-05; 2.2404455e-05; 2.2170171e-05; 2.1939410e-05; 2.1721325e-05; 2.1497868e-05; 2.1281746e-05; 2.1066392e-05; 2.0865361e-05; 2.0663689e-05; 2.0461842e-05; 2.0262440e-05; 2.0067492e-05; 1.9871680e-05; 1.9686626e-05; 1.9493360e-05; 1.9311669e-05; 1.9130045e-05; 1.8957471e-05; 1.8791214e-05; 1.8613745e-05; 1.8438778e-05; 1.8268056e-05; 1.8107355e-05; 1.7936520e-05; 1.7774794e-05; 1.7622336e-05; 1.7468209e-05; 1.7317073e-05; 1.7156560e-05; 1.7006635e-05; 1.6850353e-05; 1.6694566e-05; 1.6544633e-05; 1.6397051e-05; 1.6247227e-05; 1.6105513e-05; 1.5959169e-05; 1.5814045e-05; 1.5674216e-05; 1.5529443e-05; 1.5393004e-05; 1.5250791e-05; 1.5113162e-05; 1.4982841e-05; 1.4851411e-05; 1.4732971e-05; 1.4608880e-05; 1.4476799e-05; 1.4355466e-05; 1.4239302e-05; 1.4123552e-05; 1.4009096e-05; 1.3901341e-05; 1.3791707e-05; 1.3684883e-05; 1.3579309e-05; 1.3477250e-05; 1.3380964e-05; 1.3285072e-05; 1.3193961e-05; 1.3096009e-05; 1.3002100e-05; 1.2916068e-05; 1.2835201e-05; 1.2757285e-05; 1.2681455e-05; 1.2600026e-05; 1.2521155e-05; 1.2440337e-05; 1.2357772e-05; 1.2271727e-05; 1.2187480e-05; 1.2105919e-05; 1.2015476e-05; 1.1930051e-05; 1.1845697e-05; 1.1757021e-05; 1.1671022e-05; 1.1581546e-05; 1.1491812e-05; 1.1393272e-05; 1.1306434e-05; 1.1214786e-05; 1.1131083e-05; 1.1043223e-05; 1.0954758e-05; 1.0871692e-05; 1.0795892e-05; 1.0712152e-05; 1.0635032e-05; 1.0549327e-05; 1.0480418e-05; 1.0407391e-05; 1.0335716e-05; 1.0273188e-05; 1.0204218e-05; 1.0141881e-05; 1.0072418e-05; 1.0017353e-05; 9.9522988e-06; 9.8920386e-06; 9.8353836e-06; 9.7804867e-06; 9.7189106e-06; 9.6664451e-06; 9.6100225e-06; 9.5542382e-06; 9.5067568e-06; 9.4502030e-06; 9.3965807e-06; 9.3494798e-06; 9.2941678e-06; 9.2441359e-06; 9.1890855e-06; 9.1451930e-06; 9.0960127e-06; 9.0431590e-06; 9.0023869e-06; 8.9532183e-06; 8.9006053e-06; 8.8479745e-06; 8.8013881e-06; 8.7516577e-06; 8.7080086e-06; 8.6631816e-06; 8.6275725e-06; 8.5827615e-06; 8.5319930e-06; 8.4890763e-06; 8.4426263e-06; 8.4025275e-06; 8.3624853e-06; 8.3214680e-06; 8.2787673e-06; 8.2353137e-06; 8.1898418e-06; 8.1513096e-06; 8.1075700e-06; 8.0670340e-06; 8.0337284e-06; 8.0018920e-06; 7.9614964e-06; 7.9189033e-06; 7.8751086e-06; 7.8318080e-06; 7.7971442e-06; 7.7644203e-06; 7.7266955e-06; 7.6918049e-06; 7.6566766e-06; 7.6184604e-06; 7.5823821e-06; 7.5456248e-06; 7.5101091e-06; 7.4779536e-06; 7.4365465e-06; 7.4009704e-06; 7.3702292e-06; 7.3334310e-06; 7.3026665e-06; 7.2704401e-06; 7.2342324e-06; 7.2020315e-06; 7.1717781e-06; 7.1417039e-06; 7.1087196e-06; 7.0759939e-06; 7.0442619e-06; 7.0108076e-06; 6.9804490e-06; 6.9507917e-06; 6.9196618e-06; 6.8874639e-06; 6.8575159e-06; 6.8274854e-06; 6.7937740e-06; 6.7664240e-06; 6.7385918e-06; 6.7103144e-06; 6.6839139e-06; 6.6528982e-06; 6.6192519e-06; 6.5894894e-06; 6.5636361e-06; 6.5432500e-06; 6.5151140e-06; 6.4848429e-06; 6.4591127e-06; 6.4325141e-06; 6.4072889e-06; 6.3824919e-06; 6.3569480e-06; 6.3299259e-06; 6.3032828e-06; 6.2768862e-06; 6.2460963e-06; 6.2183054e-06; 6.1943571e-06; 6.1681769e-06; 6.1429880e-06; 6.1211247e-06; 6.0974036e-06; 6.0719190e-06; 6.0490119e-06; 6.0229826e-06; 6.0017324e-06; 5.9807377e-06; 5.9598552e-06; 5.9347902e-06; 5.9085754e-06; 5.8847357e-06; 5.8606344e-06; 5.8379997e-06; 5.8147512e-06; 5.7926182e-06; 5.7697490e-06; 5.7427965e-06; 5.7242703e-06; 5.7085791e-06; 5.6884124e-06; 5.6655528e-06; 5.6466234e-06; 5.6318049e-06; 5.6116932e-06; 5.5879545e-06; 5.5586641e-06; 5.5414344e-06; 5.5235435e-06; 5.5018881e-06; 5.4810367e-06; 5.4599734e-06; 5.4389391e-06; 5.4197413e-06; 5.4029869e-06; 5.3808201e-06; 5.3574048e-06; 5.3414192e-06; 5.3212978e-06; 5.3065866e-06; 5.2863434e-06; 5.2672235e-06; 5.2530126e-06; 5.2301732e-06; 5.2051871e-06; 5.1899138e-06; 5.1741744e-06; 5.1558493e-06; 5.1362486e-06; 5.1232243e-06; 5.1065718e-06; 5.0824148e-06; 5.0640897e-06; 5.0522214e-06; 5.0361520e-06; 5.0196322e-06; 5.0023153e-06; 4.9880881e-06; 4.9752024e-06; 4.9549908e-06; 4.9329580e-06; 4.9186035e-06; 4.9028883e-06; 4.8814447e-06; 4.8674465e-06; 4.8553488e-06; 4.8385185e-06; 4.8170606e-06; 4.8081843e-06; 4.7925153e-06; 4.7741681e-06; 4.7584930e-06; 4.7439302e-06; 4.7261146e-06; 4.7160945e-06; 4.7014378e-06; 4.6783572e-06; 4.6730794e-06; 4.6644267e-06; 4.6438774e-06; 4.6251233e-06; 4.6066256e-06; 4.6037513e-06; 4.5924186e-06; 4.5771821e-06; 4.5600603e-06; 4.5475798e-06; 4.5354375e-06; 4.5179699e-06; 4.4980502e-06; 4.4895265e-06; 4.4796477e-06; 4.4701689e-06; 4.4547061e-06; 4.4407423e-06; 4.4278466e-06; 4.4259298e-06; 4.4114941e-06; 4.3909239e-06; 4.3874107e-06; 4.3801742e-06; 4.3607308e-06; 4.3464735e-06; 4.3277092e-06; 4.3087927e-06; 4.2987068e-06; 4.2913669e-06; 4.2825334e-06; 4.2596744e-06; 4.2453210e-06; 4.2353394e-06; 4.2338215e-06; 4.2097187e-06; 4.2027028e-06; 4.1813769e-06; 4.1840738e-06; 4.1715316e-06; 4.1650562e-06; 4.1483726e-06; 4.1363057e-06; 4.1262674e-06; 4.1210987e-06; 4.1154470e-06; 4.1031037e-06; 4.0870317e-06; 4.0676579e-06; 4.0658871e-06; 4.0558148e-06; 4.0374255e-06; 4.0349678e-06; 4.0223800e-06; 4.0118927e-06; 4.0101863e-06; 3.9944410e-06; 3.9845711e-06; 3.9743385e-06; 3.9657597e-06; 3.9595417e-06; 3.9448828e-06; 3.9307256e-06; 3.9237996e-06; 3.9097773e-06; 3.8981750e-06; 3.8791772e-06; 3.8643611e-06; 3.8523381e-06; 3.8578043e-06; 3.8376057e-06; 3.8297357e-06; 3.8259295e-06; 3.8157899e-06; 3.8143885e-06; 3.8032245e-06; 3.7968654e-06; 3.7949400e-06; 3.7762555e-06; 3.7756264e-06; 3.7668899e-06; 3.7496669e-06; 3.7575407e-06; 3.7401952e-06; 3.7333318e-06; 3.7212902e-06; 3.7178668e-06; 3.6993901e-06; 3.6921005e-06; 3.7078904e-06; 3.6847263e-06; 3.6941870e-06; 3.6710189e-06; 3.6744006e-06; 3.6592385e-06; 3.6450159e-06; 3.6485661e-06; 3.6266380e-06; 3.6123063e-06; 3.6121958e-06; 3.5930239e-06; 3.5998351e-06; 3.5844144e-06; 3.5663830e-06; 3.5845028e-06; 3.5705769e-06; 3.5778739e-06; 3.5628055e-06; 3.5417021e-06; 3.5300126e-06; 3.5093355e-06; 3.5177967e-06; 3.5267650e-06; 3.5120661e-06; 3.5255759e-06; 3.4948819e-06; 3.5152608e-06; 3.5138362e-06; 3.4801024e-06; 3.4832122e-06; 3.4710983e-06; 3.4624302e-06; 3.4577179e-06; 3.4374392e-06; 3.4566345e-06; 3.4470267e-06; 3.4399787e-06; 3.4066604e-06; 3.4241760e-06; 3.4052889e-06; 3.3899351e-06; 3.4004129e-06; 3.3917439e-06; 3.3524768e-06; 3.3684439e-06; 3.3716969e-06; 3.3562779e-06; 3.3606745e-06; 3.3416025e-06; 3.3223808e-06; 3.3404936e-06; 3.3044037e-06; 3.2897197e-06; 3.3031662e-06; 3.3094725e-06; 3.2977074e-06; 3.2868120e-06; 3.2842729e-06; 3.2950959e-06; 3.2631571e-06; 3.2476967e-06; 3.2307616e-06; 3.2587591e-06; 3.2279448e-06; 3.2208862e-06; 3.2386572e-06; 3.2182713e-06; 3.2117570e-06; 3.2176295e-06; 3.2140410e-06; 3.1795497e-06; 3.1782276e-06; 3.1654086e-06; 3.1616106e-06; 3.1764321e-06; 3.1936380e-06; 3.4556000e-06; 4.1650587e-06; 4.0542791e-06; 4.0251434e-06; 3.8889238e-06; 4.0779771e-06; 4.1321333e-06; 4.1450657e-06; 4.0610685e-06; 4.0371859e-06; 3.9578491e-06; 4.0266406e-06; 4.0750958e-06; 4.0105101e-06; 4.0781046e-06; 3.9922647e-06; 3.9834000e-06; 3.9781919e-06; 3.8736959e-06; 3.8967809e-06; 3.8965521e-06; 3.9282426e-06; 3.9243136e-06; 3.9168562e-06; 3.8629234e-06; 3.8450171e-06; 3.8709735e-06; 3.8272361e-06; 3.8802648e-06; 3.9190879e-06; 3.9120333e-06; 3.9014078e-06; 3.8935040e-06; 3.8818477e-06; 3.8605834e-06; 3.8279473e-06; 3.7866780e-06; 3.8219667e-06; 3.8293368e-06; 3.8389085e-06; 3.8231175e-06; 3.7701458e-06; 3.7504643e-06; 3.7936189e-06; 3.8193361e-06; 3.7896668e-06; 3.7746720e-06; 3.7891010e-06; 3.7872205e-06; 3.8122759e-06; 3.8382564e-06; 3.8444407e-06; 3.8253565e-06; 3.8236858e-06; 3.8007222e-06; 3.8057933e-06; 3.8635335e-06; 3.8784994e-06; 3.8785939e-06; 3.8955085e-06; 3.9228687e-06; 3.9254075e-06; 3.8621325e-06; 3.7971498e-06; 3.7667713e-06; 3.7246184e-06; 3.6983877e-06; 3.6897984e-06; 3.6632219e-06; 3.6469764e-06; 3.6613463e-06; 3.6265455e-06; 3.5868416e-06; 3.5750253e-06; 3.5753516e-06; 3.6037324e-06; 3.6367727e-06; 3.6282542e-06; 3.6246225e-06; 3.6423127e-06; 3.6559496e-06; 3.6488727e-06; 3.6296475e-06; 3.6139466e-06; 3.6134050e-06; 3.6115106e-06; 3.6129889e-06; 3.6088052e-06; 3.5594360e-06; 3.5205772e-06; 3.5153836e-06; 3.5068338e-06; 3.5108342e-06; 3.5110054e-06; 3.5187196e-06; 3.5240645e-06; 3.5179042e-06; 3.5115743e-06; 3.5006712e-06; 3.5017485e-06; 3.4867069e-06; 3.4641581e-06; 3.4533276e-06; 3.4404014e-06; 3.4363372e-06; 3.4260993e-06; 3.4195630e-06; 3.4214993e-06; 3.4148271e-06; 3.4206462e-06; 3.4233680e-06; 3.4098760e-06; 3.4021443e-06; 3.4047447e-06; 3.4296207e-06; 3.4304632e-06; 3.4135771e-06; 3.3965511e-06; 3.3970896e-06; 3.3901442e-06; 3.3610977e-06; 3.3520410e-06; 3.3688109e-06; 3.3826519e-06; 3.3944758e-06; 3.3856717e-06; 3.3676740e-06; 3.3645003e-06; 3.3482768e-06; 3.3586559e-06; 3.3623346e-06; 3.3660354e-06; 3.3786358e-06; 3.3889629e-06; 3.4011275e-06; 3.4170837e-06; 3.4025219e-06; 3.4035824e-06; 3.3995415e-06; 3.3807784e-06; 3.4010567e-06; 3.4170582e-06; 3.4229854e-06; 3.4354684e-06; 3.4181589e-06; 3.4243595e-06; 3.4377462e-06; 3.4356391e-06; 3.4298729e-06; 3.4104039e-06; 3.4125583e-06; 3.4403353e-06; 3.4206939e-06; 3.3946954e-06; 3.3883386e-06; 3.3913974e-06; 3.3741870e-06; 3.3680901e-06; 3.3654086e-06; 3.3831457e-06; 3.4106980e-06; 3.4102347e-06; 3.4137524e-06; 3.3963840e-06; 3.4049490e-06; 3.3822252e-06; 3.3630202e-06; 3.3694731e-06; 3.3511207e-06; 3.3705982e-06; 3.3656977e-06; 3.3698564e-06; 3.3725026e-06; 3.3589720e-06; 3.3439557e-06; 3.3235114e-06; 3.3042666e-06; 3.3120930e-06; 3.3376791e-06; 3.3389479e-06; 3.3347192e-06; 3.3475979e-06; 3.3455352e-06; 3.3357589e-06; 3.3225820e-06; 3.3051946e-06; 3.2955176e-06; 3.3002610e-06; 3.2972593e-06; 3.2864641e-06; 3.3044245e-06; 3.2976836e-06; 3.2909204e-06; 3.2729779e-06; 3.2598698e-06; 3.2796456e-06; 3.2726606e-06; 3.2720967e-06; 3.2469141e-06; 3.2609777e-06; 3.2424170e-06; 3.2476189e-06; 3.2453796e-06; 3.2487882e-06; 3.2266650e-06; 3.2297770e-06; 3.2002872e-06; 3.1793390e-06; 3.1559199e-06; 3.1389798e-06; 3.1220144e-06; 3.1257918e-06; 3.1385326e-06; 3.1547972e-06; 3.1472285e-06; 3.1241809e-06; 3.0965284e-06; 3.1054550e-06; 3.0821521e-06; 3.0869749e-06; 3.1037897e-06; 3.1188345e-06; 3.1182613e-06; 3.1336465e-06; 3.1291902e-06; 3.1276028e-06; 3.1333019e-06; 3.1299107e-06; 3.1004118e-06; 3.0952970e-06; 3.0813114e-06; 3.1064456e-06; 3.1177763e-06; 3.1442149e-06; 3.1435992e-06; 3.1501555e-06; 3.1456626e-06; 3.1469103e-06; 3.1080560e-06; 3.1106393e-06; 3.1243703e-06; 3.1213639e-06; 3.1245100e-06; 3.1347621e-06; 3.1395089e-06; 3.1291389e-06; 3.1369825e-06; 3.1482223e-06; 3.1387491e-06; 3.1539154e-06; 3.1483065e-06; 3.1613529e-06; 3.1656484e-06; 3.1820803e-06; 3.2095570e-06; 3.2406715e-06; 3.2602733e-06; 3.3116027e-06; 3.3407280e-06; 3.3754553e-06; 3.4217368e-06; 3.4825752e-06; 3.5327898e-06; 3.6053543e-06; 3.7115674e-06; 3.8106588e-06; 3.9332080e-06; 4.0616519e-06; 4.2056980e-06; 4.3638134e-06; 4.5182788e-06; 4.7395271e-06; 4.9898607e-06; 5.2398423e-06; 5.4960638e-06; 5.7641743e-06; 6.0513802e-06; 6.3298787e-06; 6.5867339e-06; 6.8281952e-06; 7.0460273e-06; 7.2751294e-06; 7.4521421e-06; 7.5483057e-06; 7.5831976e-06; 7.5066710e-06; 7.4052113e-06; 7.2060572e-06; 6.9465219e-06; 6.7066795e-06; 6.4832642e-06; 6.2508912e-06; 6.0455819e-06; 5.8817353e-06; 5.7260126e-06; 5.5653821e-06; 5.4356375e-06; 5.2999058e-06; 5.1543684e-06; 5.0190735e-06; 4.8586291e-06; 4.7435408e-06; 4.6258222e-06; 4.5504381e-06; 4.4923809e-06; 4.4628117e-06; 4.4881155e-06; 4.5593011e-06; 4.6567164e-06; 4.7554243e-06; 4.8527799e-06; 4.9146713e-06; 4.9682512e-06; 5.0274123e-06; 5.0553027e-06; 5.0975847e-06; 5.2294022e-06; 5.4171104e-06; 5.6190863e-06; 5.8372033e-06; 5.9800981e-06; 6.0905435e-06; 6.0671204e-06; 5.9202887e-06; 5.7070638e-06; 5.4558023e-06; 5.3678179e-06; 5.3967550e-06; 5.4355885e-06; 5.5608330e-06; 5.8289276e-06; 5.6951734e-06; 5.7095455e-06; 5.7243093e-06; 5.7973817e-06; 6.0390638e-06; 6.3412507e-06; 6.6181512e-06; 6.7093185e-06; 6.7377020e-06; 6.6898104e-06; 6.4949485e-06; 6.2937984e-06; 5.9517632e-06; 5.5708092e-06; 5.2009154e-06; 4.8405064e-06; 4.5458115e-06; 4.2245240e-06; 3.9878070e-06; 3.8818287e-06; 3.4647082e-06; 3.1190551e-06; 2.8470396e-06; 2.6672693e-06; 2.5589270e-06; 2.6316248e-06; 2.8235919e-06; 2.9934940e-06; 3.2341108e-06; 3.6074002e-06; 3.6211223e-06; 3.6371228e-06; 3.6368204e-06; 3.6159492e-06; 3.5707686e-06; 3.6919704e-06; 3.7747697e-06; 3.8518726e-06; 3.9015838e-06; 3.9076828e-06; 3.9272847e-06; 3.9275724e-06; 3.9002995e-06; 3.8820641e-06; 3.8698680e-06; 3.8983665e-06; 3.8614140e-06; 3.7888917e-06; 3.7195165e-06; 3.6616298e-06; 3.6956146e-06; 3.7190360e-06; 3.6903334e-06; 3.6561381e-06; 3.6276308e-06; 3.6119055e-06; 3.5998737e-06; 3.5652399e-06; 3.5081541e-06; 3.4478377e-06; 3.4461761e-06; 3.4244217e-06; 3.3713340e-06; 3.3390384e-06; 3.3193072e-06; 3.2994491e-06; 3.2760800e-06; 3.2523525e-06; 3.2256368e-06; 3.2028883e-06; 3.1932852e-06; 3.1723612e-06; 3.1831493e-06; 3.1837886e-06; 3.1632256e-06; 3.1495386e-06; 3.1363604e-06; 3.1272319e-06; 3.1288313e-06; 3.1120124e-06; 3.1078192e-06; 3.1008193e-06; 3.0987800e-06; 3.1071706e-06; 3.1247639e-06; 3.1096505e-06; 3.0810956e-06; 3.0578495e-06; 3.0584480e-06; 3.0605145e-06; 3.0601128e-06; 3.0481234e-06; 3.0533786e-06; 3.0371429e-06; 3.0168106e-06; 3.0067328e-06; 3.0135964e-06; 3.0152073e-06; 3.0121211e-06; 3.0207704e-06; 3.0221330e-06; 3.0359628e-06; 3.0382035e-06; 3.0328267e-06; 3.0423186e-06; 3.0606496e-06; 3.0788783e-06; 3.0797641e-06; 3.0644527e-06; 3.0652287e-06; 3.0788463e-06; 3.0735051e-06; 3.0801734e-06; 3.0906837e-06; 3.0753231e-06; 3.0513176e-06; 3.0520458e-06; 3.0339375e-06; 3.0476377e-06; 3.0400682e-06; 3.0345333e-06; 3.0466720e-06; 3.0572467e-06; 3.0835489e-06; 3.0956170e-06; 3.1170950e-06; 3.1037748e-06; 3.1020656e-06; 3.1094118e-06; 3.1319306e-06; 3.1358344e-06; 3.1697284e-06; 3.1630461e-06; 3.1892136e-06; 3.2068580e-06; 3.2175824e-06; 3.2315949e-06; 3.2181311e-06; 3.2362752e-06; 3.2358902e-06; 3.2663646e-06; 3.2960856e-06; 3.3040927e-06; 3.3115061e-06; 3.2888607e-06; 3.3345165e-06; 3.3559806e-06; 3.3999031e-06; 3.4156031e-06; 3.3958205e-06; 3.4557400e-06; 3.5043888e-06; 3.5386234e-06; 3.5739551e-06; 3.5827923e-06; 3.6613209e-06; 3.6877534e-06; 3.7528497e-06; 3.7402790e-06; 3.7749218e-06; 3.8247746e-06; 3.8761548e-06; 3.8891225e-06; 3.8697094e-06; 3.7401150e-06; 3.5930763e-06; 3.5794102e-06; 3.4564346e-06; 3.4839079e-06; 3.4854438e-06; 3.5138448e-06; 3.8072975e-06; 4.0177519e-06; 4.3176731e-06; 4.5740233e-06; 4.7777388e-06; 4.9217741e-06; 5.1593263e-06; 5.1592528e-06; 5.2168185e-06; 5.3355564e-06; 5.3819755e-06; 5.5206077e-06; 5.6227430e-06; 5.6704114e-06; 5.8993722e-06; 6.1311975e-06; 6.3457858e-06; 6.6181540e-06; 6.8898246e-06; 6.8630293e-06; 6.9470850e-06; 6.7366112e-06; 6.4501206e-06; 6.0973267e-06; 5.8114746e-06; 5.3382263e-06; 4.8148631e-06; 4.6274704e-06; 4.4612460e-06; 4.2046519e-06; 4.0603362e-06; 4.1535747e-06; 4.5734667e-06; 5.1609804e-06; 5.9073668e-06; 6.4634531e-06; 6.9499698e-06; 7.3791208e-06; 7.6832150e-06; 7.8048485e-06; 7.7922105e-06; 7.8294675e-06; 7.5941035e-06; 7.5177732e-06; 7.1312698e-06; 6.7838369e-06; 6.1439055e-06; 5.7452843e-06; 5.2628844e-06; 4.9090268e-06; 4.6584005e-06; 4.5674162e-06; 4.7701995e-06; 5.0621568e-06; 5.4518765e-06; 5.8684916e-06; 6.1330441e-06; 6.4534333e-06; 6.6923288e-06; 6.8431591e-06; 6.8684081e-06; 6.8333123e-06; 6.6844439e-06; 6.5345301e-06; 6.5247184e-06; 6.3325605e-06; 6.0219911e-06; 5.7497080e-06; 5.5616207e-06; 5.3995859e-06; 5.3426775e-06; 5.2262706e-06; 5.2299589e-06; 5.2069564e-06; 5.2118919e-06; 5.2360731e-06; 5.2423143e-06; 5.3010534e-06; 5.3285343e-06; 5.2703456e-06; 5.1736117e-06; 5.1449884e-06; 5.0827261e-06; 5.0340860e-06; 5.0171453e-06; 5.0457690e-06; 5.0940733e-06; 5.1269373e-06; 5.1690336e-06; 5.2598281e-06; 5.3755263e-06; 5.4651854e-06; 5.4774763e-06; 5.4517278e-06; 5.4194443e-06; 5.3862319e-06; 5.4060733e-06; 5.3379552e-06; 5.2863680e-06; 5.2297980e-06; 5.1215749e-06; 5.0281869e-06; 4.9571150e-06; 4.8935582e-06; 4.8100777e-06; 4.7500954e-06; 4.7035576e-06; 4.6100563e-06; 4.6010371e-06; 4.5863601e-06; 4.5305953e-06; 4.5411105e-06; 4.5193213e-06; 4.5381546e-06; 4.5172716e-06; 4.4936628e-06; 4.4842483e-06; 4.4558932e-06; 4.4716739e-06; 4.4942703e-06; 4.5170836e-06; 4.5064124e-06; 4.5344546e-06; 4.5713148e-06; 4.6388453e-06; 4.6890161e-06; 4.7088232e-06; 4.7416027e-06; 4.7216265e-06; 4.6986407e-06; 4.6565370e-06; 4.5496606e-06; 4.4867019e-06; 4.4062115e-06; 4.3607815e-06; 4.3120399e-06; 4.2873033e-06; 4.2827648e-06; 4.2681748e-06; 4.2565090e-06; 4.2027623e-06; 4.1473240e-06; 4.1170731e-06; 4.0926537e-06; 4.0943970e-06; 4.1023026e-06; 4.0742482e-06; 4.0144941e-06; 3.9609928e-06; 3.9284059e-06; 3.9100731e-06; 3.8856079e-06; 3.8287816e-06; 3.8167855e-06; 3.8155430e-06; 3.7660574e-06; 3.7563833e-06; 3.7312749e-06; 3.7200653e-06; 3.7168997e-06; 3.7013533e-06; 3.7063973e-06; 3.6998369e-06; 3.7001527e-06; 3.7111938e-06; 3.7126099e-06; 3.6843695e-06; 3.6354652e-06; 3.5965221e-06; 3.5439979e-06; 3.4951276e-06; 3.5088952e-06; 3.5397291e-06; 3.5988022e-06; 3.6586479e-06; 3.6669701e-06; 3.6887681e-06; 3.7209867e-06; 3.7697197e-06; 3.8145908e-06; 3.8388717e-06; 3.8503385e-06; 3.8325769e-06; 3.8234091e-06; 3.7683978e-06; 3.7146966e-06; 3.6821718e-06; 3.6598678e-06; 3.5955548e-06; 3.5242703e-06; 3.4802660e-06; 3.4773107e-06; 3.4835525e-06; 3.4571352e-06; 3.4135682e-06; 3.3789446e-06; 3.3823837e-06; 3.3993632e-06; 3.4117522e-06; 3.4293724e-06; 3.4174571e-06; 3.4009768e-06; 3.3529206e-06; 3.3436053e-06; 3.3902891e-06; 3.4047639e-06; 3.4194802e-06; 3.4369770e-06; 3.4474892e-06; 3.4720715e-06; 3.4749971e-06; 3.4828469e-06; 3.5077172e-06; 3.5504114e-06; 3.5849157e-06; 3.5981650e-06; 3.6069937e-06; 3.5549924e-06; 3.5609069e-06; 3.5967517e-06; 3.6204916e-06; 3.6624843e-06; 3.6773271e-06; 3.6837233e-06; 3.6747474e-06; 3.6746887e-06; 3.7211401e-06; 3.7455437e-06; 3.7432641e-06; 3.7491565e-06; 3.7669886e-06; 3.7297030e-06; 3.7248205e-06; 3.7711081e-06; 3.7706518e-06; 3.7780835e-06; 3.8073200e-06; 3.8086867e-06; 3.8030010e-06; 3.8151462e-06; 3.9106499e-06; 3.9597507e-06; 4.0064598e-06; 4.0433946e-06; 3.9946061e-06; 3.9854006e-06; 4.0356824e-06; 4.1153993e-06; 4.1144398e-06; 4.2127353e-06; 4.1436669e-06; 4.2399377e-06; 4.2692129e-06; 4.3382609e-06; 4.4083803e-06; 4.4773996e-06; 4.5776905e-06; 4.6801359e-06; 4.8110565e-06; 4.9030421e-06; 5.0110616e-06; 5.1130823e-06; 5.2467474e-06; 5.2989339e-06; 5.3925753e-06; 5.5006055e-06; 5.5979762e-06; 5.6596310e-06; 5.8133633e-06; 6.2800217e-06; 6.4440653e-06; 6.6483160e-06; 6.8757075e-06; 7.1420955e-06; 7.4518045e-06; 7.7758877e-06; 8.0915410e-06; 8.3738925e-06; 8.6734935e-06; 9.0196069e-06; 9.3674229e-06; 9.7162344e-06; 1.0066036e-05; 1.0417555e-05; 1.0780726e-05; 1.1145779e-05; 1.1531029e-05; 1.1936761e-05; 1.2344788e-05; 1.3077759e-05; 1.3820702e-05; 1.4581998e-05; 1.5362053e-05; 1.6153053e-05; 1.6991025e-05; 1.7841283e-05; 1.8747502e-05; 1.9711723e-05; 2.0691606e-05; 2.2106470e-05; 2.3557429e-05; 2.5086465e-05; 2.6698366e-05; 2.8356305e-05; 3.0093422e-05; 3.1881372e-05; 3.3678083e-05; 3.5484643e-05; 3.7348801e-05; 3.7659692e-05; 3.7977859e-05; 3.8233934e-05; 3.8428104e-05; 3.8629522e-05; 3.8586715e-05; 3.8541676e-05; 3.8388886e-05; 3.8115291e-05; 3.7817678e-05; 3.7571693e-05; 3.7351497e-05; 3.7359454e-05; 3.7586625e-05; 3.7826472e-05; 4.0682348e-05; 4.3531325e-05; 4.5854030e-05; 4.7611344e-05; 4.9300470e-05; 5.0369007e-05; 5.1451615e-05; 5.2897209e-05; 5.4612511e-05; 5.6240719e-05; 6.1877248e-05; 6.7550511e-05; 7.3111360e-05; 7.8561756e-05; 8.4049279e-05; 8.3877151e-05; 8.3750980e-05; 8.3661326e-05; 8.3594159e-05; 8.3550145e-05; 8.4304308e-05; 8.5068924e-05; 8.5962150e-05; 8.6982900e-05; 8.8010956e-05; 8.9712446e-05; 9.1401922e-05; 9.3216900e-05; 9.5160800e-05; 9.7096315e-05; 9.9600735e-05; 1.0210170e-04; 1.0467978e-04; 1.0733968e-04; 1.1000218e-04; 1.0995354e-04; 1.0990873e-04; 1.0986762e-04; 1.0982740e-04; 1.0978750e-04; 1.0952816e-04; 1.0926825e-04; 1.0891270e-04; 1.0846193e-04; 1.0801112e-04; 1.0390471e-04; 9.9800672e-05; 9.5721517e-05; 9.1666247e-05; 8.7611747e-05; 8.4425645e-05; 8.1239396e-05; 7.8048817e-05; 7.4854146e-05; 7.1659538e-05; 7.0724555e-05; 6.9786547e-05; 6.8858544e-05; 6.7940508e-05; 6.7019454e-05; 6.6117783e-05; 6.5214123e-05; 6.4302050e-05; 6.3381848e-05; 6.2460047e-05; 6.2812642e-05; 6.3166177e-05; 6.3501957e-05; 6.3820105e-05; 6.4139348e-05; 6.4389748e-05; 6.4640266e-05; 6.4878620e-05; 6.5104753e-05; 6.5330849e-05; 6.3538141e-05; 6.1743753e-05; 5.9964901e-05; 5.8202537e-05; 5.6439682e-05; 5.4863615e-05; 5.3287810e-05; 5.1748993e-05; 5.0246698e-05; 4.8744019e-05; 4.6714115e-05; 4.4680719e-05; 4.2653363e-05; 4.0631951e-05; 3.8606940e-05; 3.6885203e-05; 3.5160184e-05; 3.3428967e-05; 3.1691560e-05; 2.9950905e-05; 2.9291763e-05; 2.8630905e-05; 2.7960243e-05; 2.7279876e-05; 2.6597944e-05; 2.5837634e-05; 2.5075801e-05; 2.4303118e-05; 2.3519662e-05; 2.2734793e-05; 2.2628730e-05; 2.2522879e-05; 2.2413435e-05; 2.2300407e-05; 2.2187600e-05; 2.2073741e-05; 2.1960074e-05; 2.1850808e-05; 2.1745914e-05; 2.1641172e-05; 2.1638856e-05; 2.1636617e-05; 2.1636542e-05; 2.1638614e-05; 2.1640745e-05; 2.1651507e-05; 2.1662321e-05; 2.1673937e-05; 2.1686354e-05; 2.1698818e-05; 2.1780370e-05; 2.1862116e-05; 2.1947102e-05; 2.2035345e-05; 2.2123809e-05; 2.2223903e-05; 2.2324276e-05; 2.2433405e-05; 2.2551360e-05; 2.2669691e-05; 2.2709963e-05; 2.2751918e-05; 2.2800620e-05; 2.2856179e-05; 2.2913601e-05; 2.3006615e-05; 2.3101758e-05; 2.3204402e-05; 2.3314708e-05; 2.3427404e-05; 2.3370478e-05; 2.3316759e-05; 2.3267324e-05; 2.3222242e-05; 2.3180542e-05; 2.3231853e-05; 2.3286821e-05; 2.3343014e-05; 2.3400428e-05; 2.3461587e-05; 2.3254222e-05; 2.3048962e-05; 2.2838657e-05; 2.2623099e-05; 2.2409419e-05; 2.2331960e-05; 2.2256497e-05; 2.2204196e-05; 2.2175108e-05; 2.2148127e-05; 2.2018545e-05; 2.1889723e-05; 2.1754156e-05; 2.1611904e-05; 2.1470513e-05; 2.1397023e-05; 2.1324454e-05; 2.1242329e-05; 2.1150573e-05; 2.1059653e-05; 2.0956147e-05; 2.0852740e-05; 2.0745611e-05; 2.0634754e-05; 2.0523994e-05; 2.0425195e-05; 2.0326515e-05; 2.0222815e-05; 2.0114086e-05; 2.0005465e-05; 1.9971933e-05; 1.9938346e-05; 1.9902090e-05; 1.9863154e-05; 1.9824148e-05; 1.9776063e-05; 1.9727898e-05; 1.9680304e-05; 1.9633282e-05; 1.9586183e-05; 1.9623618e-05; 1.9661056e-05; 1.9703165e-05; 1.9749941e-05; 1.9796718e-05; 1.9844086e-05; 1.9891461e-05; 1.9939313e-05; 1.9987645e-05; 2.0035984e-05; 2.0030444e-05; 2.0024880e-05; 2.0015577e-05; 2.0002533e-05; 1.9989463e-05; 1.9965194e-05; 1.9940894e-05; 1.9915191e-05; 1.9888080e-05; 1.9860930e-05; 1.9750797e-05; 1.9640542e-05; 1.9522789e-05; 1.9397528e-05; 1.9272131e-05; 1.9158020e-05; 1.9043780e-05; 1.8931386e-05; 1.8820849e-05; 1.8710195e-05; 1.8671252e-05; 1.8632257e-05; 1.8604009e-05; 1.8586536e-05; 1.8569047e-05; 1.8554785e-05; 1.8540519e-05; 1.8532091e-05; 1.8529525e-05; 1.8526991e-05; 1.8518583e-05; 1.8510208e-05; 1.8496904e-05; 1.8478644e-05; 1.8460384e-05; 1.8436503e-05; 1.8412599e-05; 1.8385094e-05; 1.8353974e-05; 1.8322815e-05; 1.8238731e-05; 1.8154564e-05; 1.8069437e-05; 1.7983351e-05; 1.7897182e-05; 1.7810836e-05; 1.7724412e-05; 1.7641450e-05; 1.7561955e-05; 1.7482390e-05; 1.7427422e-05; 1.7372419e-05; 1.7317095e-05; 1.7261451e-05; 1.7205773e-05; 1.7154589e-05; 1.7103368e-05; 1.7046925e-05; 1.6985263e-05; 1.6923568e-05; 1.6892813e-05; 1.6862034e-05; 1.6828362e-05; 1.6791794e-05; 1.6755197e-05; 1.6720610e-05; 1.6685997e-05; 1.6652511e-05; 1.6620151e-05; 1.6587764e-05; 1.6627406e-05; 1.6667121e-05; 1.6708837e-05; 1.6752559e-05; 1.6796362e-05; 1.6862017e-05; 1.6927804e-05; 1.6996347e-05; 1.7067656e-05; 1.7139111e-05; 1.7254228e-05; 1.7369642e-05; 1.7490901e-05; 1.7618020e-05; 1.7745461e-05; 1.7868527e-05; 1.7991910e-05; 1.8110845e-05; 1.8225309e-05; 1.8340062e-05; 1.8369235e-05; 1.8398426e-05; 1.8421571e-05; 1.8438660e-05; 1.8455754e-05; 1.8465762e-05; 1.8475770e-05; 1.8477270e-05; 1.8470250e-05; 1.8463208e-05; 1.8433569e-05; 1.8403911e-05; 1.8377306e-05; 1.8353758e-05; 1.8330196e-05; 1.8312616e-05; 1.8295027e-05; 1.8282770e-05; 1.8275846e-05; 1.8268913e-05; 1.8314497e-05; 1.8360247e-05; 1.8409040e-05; 1.8460895e-05; 1.8512946e-05; 1.8576908e-05; 1.8641114e-05; 1.8711966e-05; 1.8789497e-05; 1.8867313e-05; 1.8981464e-05; 1.9096285e-05; 1.9215510e-05; 1.9339170e-05; 1.9463542e-05; 1.9580951e-05; 1.9699044e-05; 1.9812835e-05; 1.9922306e-05; 2.0032440e-05; 2.0034603e-05; 2.0036869e-05; 2.0032499e-05; 2.0021458e-05; 2.0010473e-05; 1.9992878e-05; 1.9975306e-05; 1.9952205e-05; 1.9923557e-05; 1.9894906e-05; 1.9809140e-05; 1.9723306e-05; 1.9636275e-05; 1.9548048e-05; 1.9459753e-05; 1.9376730e-05; 1.9293640e-05; 1.9210277e-05; 1.9126641e-05; 1.9042936e-05; 1.8968211e-05; 1.8893429e-05; 1.8817543e-05; 1.8740555e-05; 1.8663514e-05; 1.8587434e-05; 1.8511300e-05; 1.8439976e-05; 1.8373459e-05; 1.8306883e-05; 1.8281452e-05; 1.8256012e-05; 1.8234959e-05; 1.8218307e-05; 1.8201668e-05; 1.8190986e-05; 1.8180325e-05; 1.8168865e-05; 1.8156605e-05; 1.8144360e-05; 1.8182765e-05; 1.8221334e-05; 1.8260515e-05; 1.8300314e-05; 1.8340284e-05; 1.8390867e-05; 1.8441643e-05; 1.8493334e-05; 1.8545947e-05; 1.8598763e-05; 1.8685362e-05; 1.8772233e-05; 1.8855113e-05; 1.8933974e-05; 1.9013074e-05; 1.9093372e-05; 1.9173907e-05; 1.9256106e-05; 1.9339986e-05; 1.9424126e-05; 1.9536740e-05; 1.9649711e-05; 1.9759413e-05; 1.9865824e-05; 1.9972560e-05; 2.0082710e-05; 2.0193200e-05; 2.0306749e-05; 2.0423373e-05; 2.0540359e-05; 2.0653863e-05; 2.0767714e-05; 2.0880939e-05; 2.0993530e-05; 2.1106461e-05; 2.1221145e-05; 2.1336178e-05; 2.1450265e-05; 2.1563406e-05; 2.1676894e-05; 2.1968083e-05; 2.2260695e-05; 2.2564551e-05; 2.2879775e-05; 2.3196604e-05; 2.3539023e-05; 2.3883292e-05; 2.4239701e-05; 2.4608416e-05; 2.4979229e-05; 2.5561576e-05; 2.6148498e-05; 2.6746712e-05; 2.7356407e-05; 2.7971037e-05; 2.8589373e-05; 2.9212716e-05; 2.9837498e-05; 3.0463711e-05; 3.1095035e-05; 3.1517561e-05; 3.1945863e-05; 3.2382004e-05; 3.2826152e-05; 3.3276542e-05; 3.3771512e-05; 3.4273407e-05; 3.4800890e-05; 3.5354615e-05; 3.5916458e-05; 3.6946327e-05; 3.8000049e-05; 3.9099339e-05; 4.0246195e-05; 4.1421433e-05; 4.2665746e-05; 4.3941829e-05; 4.5260877e-05; 4.6625111e-05; 4.8026828e-05; 5.0209724e-05; 5.2450860e-05; 5.4740835e-05; 5.7081861e-05; 5.9489465e-05; 6.1944355e-05; 6.4472139e-05; 6.7062346e-05; 6.9718407e-05; 7.2459615e-05; 7.6048494e-05; 7.9705532e-05; 8.3441979e-05; 8.7263815e-05; 9.1170807e-05; 9.5168447e-05; 9.9263354e-05; 1.0342870e-04; 1.0766812e-04; 1.1202334e-04; 1.1439797e-04; 1.1680057e-04; 1.1915843e-04; 1.2146786e-04; 1.2380216e-04; 1.2604588e-04; 1.2831069e-04; 1.3049924e-04; 1.3260603e-04; 1.3472772e-04; 1.3366891e-04; 1.3263233e-04; 1.3153733e-04; 1.3038675e-04; 1.2926092e-04; 1.2806345e-04; 1.2689320e-04; 1.2569271e-04; 1.2446387e-04; 1.2326310e-04; 1.2002459e-04; 1.1692070e-04; 1.1388490e-04; 1.1091434e-04; 1.0805898e-04; 1.0524029e-04; 1.0252796e-04; 9.9875229e-05; 9.7279621e-05; 9.4776062e-05; 9.1951307e-05; 8.9229904e-05; 8.6601432e-05; 8.4060174e-05; 8.1604399e-05; 7.9285327e-05; 7.7040526e-05; 7.4896520e-05; 7.2847149e-05; 6.7616916e-05; 6.7207501e-05; 6.6798017e-05; 6.6388465e-05; 6.5978844e-05; 6.5569155e-05; 6.5159397e-05; 6.4749571e-05; 6.4339676e-05; 6.3929713e-05; 6.3519681e-05; 6.4161042e-05; 6.4803281e-05; 6.5446399e-05; 6.6090396e-05; 6.6735273e-05; 6.7381032e-05; 6.8027672e-05; 6.8675196e-05; 6.9323603e-05; 6.9972896e-05; 6.9308886e-05; 6.8644765e-05; 6.7980534e-05; 6.7316193e-05; 6.6651740e-05; 6.5987176e-05; 6.5322502e-05; 6.4657716e-05; 6.3992818e-05; 6.3327809e-05; 6.3602233e-05; 6.3876992e-05; 6.4152087e-05; 6.4427517e-05; 6.4703283e-05; 6.4979385e-05; 6.5255824e-05; 6.5532600e-05; 6.5809713e-05; 6.6087162e-05; 6.6749734e-05; 6.7413716e-05; 6.8079111e-05; 6.8745922e-05; 6.9414152e-05; 7.0083805e-05; 7.0754883e-05; 7.1427389e-05; 7.2101327e-05; 7.2776699e-05; 7.3756268e-05; 7.4737510e-05; 7.5720429e-05; 7.6705028e-05; 7.7691310e-05; 7.8679278e-05; 7.9668935e-05; 8.0660283e-05; 8.1653327e-05; 8.2648069e-05; 8.6671234e-05; 9.0715424e-05; 9.4780827e-05; 9.8867631e-05; 1.0297603e-04; 1.0710621e-04; 1.1125838e-04; 1.1543274e-04; 1.1962948e-04; 1.2384881e-04; 1.2999060e-04; 1.3618150e-04; 1.4242222e-04; 1.4871347e-04; 1.5505598e-04; 1.6145052e-04; 1.6789783e-04; 1.7439872e-04; 1.8095398e-04; 1.8756443e-04; 1.9141183e-04; 1.9527966e-04; 1.9916811e-04; 2.0307736e-04; 2.0700759e-04; 2.1095899e-04; 2.1493175e-04; 2.1892606e-04; 2.2294212e-04; 2.2698012e-04; 2.2634369e-04; 2.2570711e-04; 2.2507040e-04; 2.2443354e-04; 2.2379654e-04; 2.2315939e-04; 2.2252211e-04; 2.2188468e-04; 2.2124711e-04; 2.2060940e-04; 2.1786930e-04; 2.1513683e-04; 2.1241192e-04; 2.0969452e-04; 2.0698457e-04; 2.0428201e-04; 2.0158680e-04; 1.9889888e-04; 1.9621820e-04; 1.9354470e-04; 1.9107201e-04; 1.8860434e-04; 1.8614165e-04; 1.8368392e-04; 1.8123110e-04; 1.7878318e-04; 1.7634010e-04; 1.7390186e-04; 1.7146840e-04; 1.6903970e-04; 1.7303433e-04; 1.7705042e-04; 1.8108818e-04; 1.8514779e-04; 1.8922946e-04; 1.9333338e-04; 1.9745976e-04; 2.0160881e-04; 2.0578073e-04; 2.0997574e-04; 2.1901428e-04; 2.2815174e-04; 2.3739018e-04; 2.4673170e-04; 2.5617847e-04; 2.6573274e-04; 2.7539684e-04; 2.8517315e-04; 2.9506417e-04; 3.0507244e-04; 3.0625485e-04; 3.0743967e-04; 3.0862692e-04; 3.0981660e-04; 3.1100871e-04; 3.1220327e-04; 3.1340028e-04; 3.1459975e-04; 3.1580168e-04; 3.1700607e-04; 3.1336825e-04; 3.0974452e-04; 3.0613477e-04; 3.0253885e-04; 2.9895664e-04; 2.9538801e-04; 2.9183283e-04; 2.8829098e-04; 2.8476234e-04; 2.8124678e-04; 2.8383155e-04; 2.8642777e-04; 2.8903551e-04; 2.9165486e-04; 2.9428588e-04; 2.9692867e-04; 2.9958330e-04; 3.0224984e-04; 3.0492839e-04; 3.0761901e-04; 3.0554837e-04; 3.0348233e-04; 3.0142085e-04; 2.9936391e-04; 2.9731148e-04; 2.9526354e-04; 2.9322006e-04; 2.9118101e-04; 2.8914637e-04; 2.8711611e-04; 2.7867660e-04; 2.7031164e-04; 2.6201975e-04; 2.5379950e-04; 2.4564952e-04; 2.3756845e-04; 2.2955499e-04; 2.2160788e-04; 2.1372586e-04; 2.0590774e-04; 2.0272593e-04; 1.9955460e-04; 1.9639365e-04; 1.9324300e-04; 1.9010257e-04; 1.8697227e-04; 1.8385202e-04; 1.8074174e-04; 1.7764134e-04; 1.7455076e-04; 1.7520807e-04; 1.7586596e-04; 1.7652441e-04; 1.7718342e-04; 1.7784300e-04; 1.7850314e-04; 1.7916385e-04; 1.7982512e-04; 1.8048697e-04; 1.8114937e-04; 1.8384004e-04; 1.8654104e-04; 1.8925242e-04; 1.9197424e-04; 1.9470658e-04; 1.9744948e-04; 2.0020302e-04; 2.0296725e-04; 2.0574224e-04; 2.0852806e-04; 2.1513785e-04; 2.2180152e-04; 2.2851984e-04; 2.3529363e-04; 2.4212371e-04; 2.4901092e-04; 2.5595612e-04; 2.6296021e-04; 2.7002408e-04; 2.7714866e-04; 2.8414533e-04; 2.9120348e-04; 2.9832407e-04; 3.0550811e-04; 3.1275660e-04; 3.2007059e-04; 3.2745114e-04; 3.3489934e-04; 3.4241631e-04; 3.5000321e-04; 3.4111289e-04; 3.3231118e-04; 3.2359621e-04; 3.1496619e-04; 3.0641939e-04; 2.9795412e-04; 2.8956873e-04; 2.8126164e-04; 2.7303130e-04; 2.6487622e-04; 2.5706969e-04; 2.4932677e-04; 2.4164631e-04; 2.3402720e-04; 2.2646836e-04; 2.1896874e-04; 2.1152731e-04; 2.0414306e-04; 1.9681503e-04; 1.8954226e-04; 1.8732037e-04; 1.8510216e-04; 1.8288760e-04; 1.8067669e-04; 1.7846938e-04; 1.7626568e-04; 1.7406555e-04; 1.7186897e-04; 1.6967593e-04; 1.6748641e-04; 1.6647365e-04; 1.6546129e-04; 1.6444934e-04; 1.6343779e-04; 1.6242663e-04; 1.6141588e-04; 1.6040552e-04; 1.5939555e-04; 1.5838598e-04; 1.5737680e-04; 1.5551263e-04; 1.5365004e-04; 1.5178902e-04; 1.4992955e-04; 1.4807163e-04; 1.4621524e-04; 1.4436039e-04; 1.4250706e-04; 1.4065524e-04; 1.3880492e-04; 1.3871396e-04; 1.3862290e-04; 1.3853174e-04; 1.3844048e-04; 1.3834913e-04; 1.3825767e-04; 1.3816612e-04; 1.3807448e-04; 1.3798273e-04; 1.3789089e-04; 1.4021573e-04; 1.4254698e-04; 1.4488466e-04; 1.4722880e-04; 1.4957943e-04; 1.5193658e-04; 1.5430026e-04; 1.5667052e-04; 1.5904737e-04; 1.6143085e-04; 1.6668463e-04; 1.7196881e-04; 1.7728372e-04; 1.8262968e-04; 1.8800701e-04; 1.9341605e-04; 1.9885715e-04; 2.0433064e-04; 2.0983687e-04; 2.1537622e-04; 2.2118091e-04; 2.2702361e-04; 2.3290477e-04; 2.3882486e-04; 2.4478433e-04; 2.5078366e-04; 2.5682335e-04; 2.6290388e-04; 2.6902576e-04; 2.7518950e-04; 2.7802429e-04; 2.8086574e-04; 2.8371386e-04; 2.8656871e-04; 2.8943029e-04; 2.9229865e-04; 2.9517381e-04; 2.9805580e-04; 3.0094466e-04; 3.0384041e-04; 3.0156944e-04; 2.9930599e-04; 2.9704999e-04; 2.9480141e-04; 2.9256018e-04; 2.9032626e-04; 2.8809960e-04; 2.8588014e-04; 2.8366784e-04; 2.8146266e-04; 2.7875544e-04; 2.7606955e-04; 2.7340469e-04; 2.7076060e-04; 2.6813700e-04; 2.6553362e-04; 2.6295019e-04; 2.6038647e-04; 2.5784220e-04; 2.5531713e-04; 2.5161161e-04; 2.4793511e-04; 2.4428725e-04; 2.4066762e-04; 2.3707583e-04; 2.3351150e-04; 2.2997426e-04; 2.2646375e-04; 2.2297960e-04; 2.1952145e-04; 2.1896599e-04; 2.1842478e-04; 2.1789732e-04; 2.1738314e-04; 2.1688181e-04; 2.1639288e-04; 2.1591597e-04; 2.1545066e-04; 2.1499660e-04; 2.1455343e-04; 2.1353082e-04; 2.1250358e-04; 2.1147167e-04; 2.1043506e-04; 2.0939370e-04; 2.0834754e-04; 2.0729656e-04; 2.0624070e-04; 2.0517992e-04; 2.0411418e-04; 2.0233556e-04; 2.0055901e-04; 1.9878454e-04; 1.9701211e-04; 1.9524173e-04; 1.9347337e-04; 1.9170702e-04; 1.8994268e-04; 1.8818032e-04; 1.8641994e-04; 1.8334555e-04; 1.8023916e-04; 1.7710026e-04; 1.7392836e-04; 1.7072293e-04; 1.6748347e-04; 1.6420943e-04; 1.6090026e-04; 1.5755540e-04; 1.5417427e-04; 1.5316049e-04; 1.5213392e-04; 1.5109434e-04; 1.5004152e-04; 1.4897525e-04; 1.4789528e-04; 1.4680138e-04; 1.4569331e-04; 1.4457082e-04; 1.4343366e-04; 1.4546019e-04; 1.4746155e-04; 1.4943825e-04; 1.5139080e-04; 1.5331968e-04; 1.5522537e-04; 1.5710831e-04; 1.5896897e-04; 1.6080777e-04; 1.6262515e-04; 1.6370062e-04; 1.6479058e-04; 1.6589532e-04; 1.6701516e-04; 1.6815041e-04; 1.6930139e-04; 1.7046844e-04; 1.7165190e-04; 1.7285212e-04; 1.7406946e-04; 1.7425487e-04; 1.7444110e-04; 1.7462817e-04; 1.7481609e-04; 1.7500485e-04; 1.7519448e-04; 1.7538497e-04; 1.7557634e-04; 1.7576859e-04; 1.7596174e-04; 1.7724984e-04; 1.7853452e-04; 1.7981582e-04; 1.8109372e-04; 1.8236824e-04; 1.8363940e-04; 1.8490720e-04; 1.8617165e-04; 1.8743276e-04; 1.8869054e-04; 1.8847361e-04; 1.8825259e-04; 1.8802746e-04; 1.8779817e-04; 1.8756469e-04; 1.8732698e-04; 1.8708500e-04; 1.8683872e-04; 1.8658809e-04; 1.8633308e-04; 1.8782617e-04; 1.8930041e-04; 1.9075620e-04; 1.9219391e-04; 1.9361393e-04; 1.9501661e-04; 1.9640231e-04; 1.9777137e-04; 1.9912413e-04; 2.0046091e-04; 2.0461626e-04; 2.0876622e-04; 2.1291082e-04; 2.1705014e-04; 2.2118421e-04; 2.2531308e-04; 2.2943682e-04; 2.3355546e-04; 2.3766907e-04; 2.4177768e-04; 2.4010214e-04; 2.3840046e-04; 2.3667209e-04; 2.3491645e-04; 2.3313296e-04; 2.3132100e-04; 2.2947995e-04; 2.2760916e-04; 2.2570797e-04; 2.2377570e-04; 2.2115684e-04; 2.1852812e-04; 2.1588950e-04; 2.1324092e-04; 2.1058232e-04; 2.0791363e-04; 2.0523480e-04; 2.0254578e-04; 1.9984648e-04; 1.9713687e-04; 1.9306917e-04; 1.8903561e-04; 1.8503565e-04; 1.8106881e-04; 1.7713458e-04; 1.7323248e-04; 1.6936204e-04; 1.6552280e-04; 1.6171431e-04; 1.5793612e-04; 1.5508456e-04; 1.5219911e-04; 1.4927922e-04; 1.4632430e-04; 1.4333377e-04; 1.4030703e-04; 1.3724346e-04; 1.3414244e-04; 1.3100332e-04; 1.2782545e-04; 1.2601162e-04; 1.2421126e-04; 1.2242419e-04; 1.2065024e-04; 1.1888926e-04; 1.1714109e-04; 1.1540556e-04; 1.1368253e-04; 1.1197184e-04; 1.1027334e-04; 1.1266999e-04; 1.1510001e-04; 1.1756408e-04; 1.2006290e-04; 1.2259721e-04; 1.2516774e-04; 1.2777526e-04; 1.3042055e-04; 1.3310441e-04; 1.3582769e-04; 1.3215055e-04; 1.2850387e-04; 1.2488732e-04; 1.2130056e-04; 1.1774325e-04; 1.1421506e-04; 1.1071567e-04; 1.0724474e-04; 1.0380195e-04; 1.0038696e-04; 1.0034242e-04; 1.0024846e-04; 1.0010457e-04; 9.9910266e-05; 9.9665024e-05; 9.9368331e-05; 9.9019661e-05; 9.8618478e-05; 9.8164236e-05; 9.7656384e-05; 1.0076160e-04; 1.0385857e-04; 1.0694740e-04; 1.1002818e-04; 1.1310101e-04; 1.1616599e-04; 1.1922321e-04; 1.2227276e-04; 1.2531473e-04; 1.2834922e-04; 1.2784921e-04; 1.2734965e-04; 1.2685055e-04; 1.2635190e-04; 1.2585370e-04; 1.2535594e-04; 1.2485863e-04; 1.2436177e-04; 1.2386534e-04; 1.2336936e-04; 1.1966369e-04; 1.1608240e-04; 1.1261925e-04; 1.0926841e-04; 1.0602444e-04; 1.0288225e-04; 9.9837040e-05; 9.6884342e-05; 9.4019941e-05; 9.1239881e-05; 8.9931126e-05; 8.8586888e-05; 8.7205777e-05; 8.5786333e-05; 8.4327017e-05; 8.2826204e-05; 8.1282182e-05; 7.9693143e-05; 7.8057176e-05; 7.6372261e-05; 7.7098402e-05; 7.7827372e-05; 7.8559185e-05; 7.9293855e-05; 8.0031396e-05; 8.0771822e-05; 8.1515146e-05; 8.2261383e-05; 8.3010546e-05; 8.3762650e-05; 8.2553347e-05; 8.1341027e-05; 8.0125678e-05; 7.8907288e-05; 7.7685844e-05; 7.6461335e-05; 7.5233748e-05; 7.4003070e-05; 7.2769288e-05; 7.1532391e-05; 7.4004689e-05; 7.6479379e-05; 7.8956479e-05; 8.1436010e-05; 8.3917990e-05; 8.6402439e-05; 8.8889376e-05; 9.1378821e-05; 9.3870792e-05; 9.6365310e-05; 9.8411475e-05; 1.0045937e-04; 1.0250901e-04; 1.0456038e-04; 1.0661350e-04; 1.0866837e-04; 1.1072500e-04; 1.1278338e-04; 1.1484353e-04; 1.1690544e-04; 1.1765260e-04; 1.1840379e-04; 1.1915904e-04; 1.1991839e-04; 1.2068187e-04; 1.2144952e-04; 1.2222136e-04; 1.2299744e-04; 1.2377779e-04; 1.2456245e-04; 1.2852723e-04; 1.3249273e-04; 1.3645902e-04; 1.4042618e-04; 1.4439426e-04; 1.4836332e-04; 1.5233345e-04; 1.5630470e-04; 1.6027714e-04; 1.6425084e-04; 1.6164429e-04; 1.5902820e-04; 1.5640253e-04; 1.5376722e-04; 1.5112222e-04; 1.4846747e-04; 1.4580293e-04; 1.4312853e-04; 1.4044423e-04; 1.3774996e-04; 1.3492146e-04; 1.3209917e-04; 1.2928306e-04; 1.2647309e-04; 1.2366923e-04; 1.2087147e-04; 1.1807977e-04; 1.1529411e-04; 1.1251445e-04; 1.0974077e-04; 1.0640565e-04; 1.0307623e-04; 9.9752470e-05; 9.6434348e-05; 9.3121828e-05; 8.9814877e-05; 8.6513464e-05; 8.3217557e-05; 7.9927125e-05; 7.6642136e-05; 7.9376941e-05; 8.2112576e-05; 8.4849077e-05; 8.7586482e-05; 9.0324829e-05; 9.3064153e-05; 9.5804492e-05; 9.8545884e-05; 1.0128837e-04; 1.0403198e-04; 1.0699202e-04; 1.0994335e-04; 1.1288602e-04; 1.1582009e-04; 1.1874563e-04; 1.2166269e-04; 1.2457134e-04; 1.2747164e-04; 1.3036364e-04; 1.3324740e-04; 1.3527270e-04; 1.3731230e-04; 1.3936634e-04; 1.4143500e-04; 1.4351841e-04; 1.4561675e-04; 1.4773017e-04; 1.4985883e-04; 1.5200291e-04; 1.5416258e-04; 1.5286170e-04; 1.5156666e-04; 1.5027739e-04; 1.4899387e-04; 1.4771604e-04; 1.4644386e-04; 1.4517729e-04; 1.4391629e-04; 1.4266081e-04; 1.4141082e-04; 1.3560453e-04; 1.2977473e-04; 1.2392109e-04; 1.1804330e-04; 1.1214104e-04; 1.0621398e-04; 1.0026179e-04; 9.4284115e-05; 8.8280617e-05; 8.2250940e-05; 8.5399695e-05; 8.8524909e-05; 9.1627057e-05; 9.4706601e-05; 9.7763989e-05; 1.0079966e-04; 1.0381403e-04; 1.0680752e-04; 1.0978053e-04; 1.1273345e-04; 1.1316099e-04; 1.1358844e-04; 1.1401579e-04; 1.1444304e-04; 1.1487018e-04; 1.1529722e-04; 1.1572416e-04; 1.1615100e-04; 1.1657773e-04; 1.1700435e-04; 1.1934797e-04; 1.2170086e-04; 1.2406309e-04; 1.2643470e-04; 1.2881575e-04; 1.3120629e-04; 1.3360636e-04; 1.3601601e-04; 1.3843530e-04; 1.4086429e-04; 1.4308662e-04; 1.4531534e-04; 1.4755031e-04; 1.4979138e-04; 1.5203840e-04; 1.5429120e-04; 1.5654960e-04; 1.5881344e-04; 1.6108251e-04; 1.6335663e-04; 1.5622909e-04; 1.4910107e-04; 1.4197240e-04; 1.3484288e-04; 1.2771236e-04; 1.2058065e-04; 1.1344757e-04; 1.0631296e-04; 9.9176635e-05; 9.2038417e-05; 9.4608411e-05; 9.7139480e-05; 9.9632533e-05; 1.0208845e-04; 1.0450808e-04; 1.0689226e-04; 1.0924180e-04; 1.1155746e-04; 1.1384002e-04; 1.1609021e-04; 1.1504968e-04; 1.1395879e-04; 1.1281744e-04; 1.1162552e-04; 1.1038293e-04; 1.0908955e-04; 1.0774525e-04; 1.0634989e-04; 1.0490334e-04; 1.0340546e-04; 1.1016552e-04; 1.1701909e-04; 1.2396902e-04; 1.3101829e-04; 1.3816995e-04; 1.4542718e-04; 1.5279328e-04; 1.6027168e-04; 1.6786590e-04; 1.7557965e-04; 1.6756767e-04; 1.5967798e-04; 1.5190747e-04; 1.4425314e-04; 1.3671212e-04; 1.2928161e-04; 1.2195892e-04; 1.1474145e-04; 1.0762669e-04; 1.0061220e-04; 1.0595497e-04; 1.1139010e-04; 1.1692052e-04; 1.2254935e-04; 1.2827981e-04; 1.3411533e-04; 1.4005948e-04; 1.4611603e-04; 1.5228894e-04; 1.5858239e-04; 1.4967626e-04; 1.4093217e-04; 1.3234244e-04; 1.2389991e-04; 1.1559782e-04; 1.0742981e-04; 9.9389925e-05; 9.1472522e-05; 8.3672294e-05; 7.5984229e-05; 9.0696536e-05; 1.0572619e-04; 1.2108473e-04; 1.3678433e-04; 1.5283783e-04; 1.6925878e-04; 1.8606151e-04; 2.0326117e-04; 2.2087382e-04; 2.3891649e-04; 2.4033228e-04; 2.4173240e-04; 2.4311704e-04; 2.4448636e-04; 2.4584052e-04; 2.4717969e-04; 2.4850403e-04; 2.4981370e-04; 2.5110887e-04; 2.5238968e-04; 2.4151076e-04; 2.3081968e-04; 2.2031080e-04; 2.0997872e-04; 1.9981829e-04; 1.8982455e-04; 1.7999276e-04; 1.7031838e-04; 1.6079703e-04; 1.5142451e-04; 1.5902137e-04; 1.6673533e-04; 1.7456921e-04; 1.8252595e-04; 1.9060856e-04; 1.9882016e-04; 2.0716401e-04; 2.1564345e-04; 2.2426197e-04];

	n_itp = interpolate((λrefn,), nref, Gridded(Linear()));
	n_itp = extrapolate(n_itp, Flat())
	k_itp = interpolate((λrefk,), kref, Gridded(Linear()));
	k_itp = extrapolate(k_itp, Flat())

	f(λ) = n_itp(λ) + k_itp(λ) * im

	return JolabFunction1D{T,Complex{T}}(f)
end
